"""
    Swagger Generator
    ~~~~~~~~~~~~~~~~~

    Base classes for swagger generators.

    :copyright: Copyright 2019 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

import abc
import functools
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING

from marshmallow import Schema

from flask_rebar.swagger_generation import swagger_words as sw
from flask_rebar.swagger_generation.authenticator_to_swagger import (
    authenticator_converter_registry as global_authenticator_converter_registry,
)
from flask_rebar.swagger_generation.authenticator_to_swagger import (
    AuthenticatorConverter,
    AuthenticatorConverterRegistry,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import (
    headers_converter_registry as global_headers_converter_registry,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import (
    query_string_converter_registry as global_query_string_converter_registry,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import (
    request_body_converter_registry as global_request_body_converter_registry,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import (
    response_converter_registry as global_response_converter_registry,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import ConverterRegistry
from flask_rebar.validation import Error


# avoid circular imports
if TYPE_CHECKING:
    from flask_rebar.rebar import HandlerRegistry


class SwaggerGeneratorI(abc.ABC):
    @abc.abstractmethod
    def get_open_api_version(self) -> str:
        """
        Rebar supports multiple OpenAPI specifications.
        :return: The OpenAPI specification the generator supports.
        """

    @abc.abstractmethod
    def generate_swagger(
        self, registry: "HandlerRegistry", host: Optional[str] = None
    ) -> Dict[str, str]:
        """
        Generate a swagger definition json object.
        :param registry:
        :param host:
        :return:
        """

    @abc.abstractmethod
    def register_flask_converter_to_swagger_type(
        self, flask_converter: str, swagger_type: Any
    ) -> None:
        """
        Flask has "converters" that convert path arguments to a Python type.

        We need to map these to Swagger types. This allows additional flask
        converter types (they're pluggable!) to be mapped to Swagger types.

        Unknown Flask converters will default to string.

        :param str flask_converter:
        :param object swagger_type:
        """


class SwaggerGenerator(SwaggerGeneratorI):
    """Base class for SwaggerV2Generator and SwaggerV3Generator.

    Inheritance is a fragile way to share code, but its a convenient one...

    :param int openapi_major_version: Major version of the Swagger specification this will produce
    :param str version: Version of the API this swagger is specifying
    :param str title: Title of the API this swagger is specifying
    :param str description: Descrption of the API this swagger is specifying

    :param ConverterRegistry query_string_converter_registry:
    :param ConverterRegistry request_body_converter_registry:
    :param ConverterRegistry headers_converter_registry:
    :param ConverterRegistry response_converter_registry:
        ConverterRegistry instances that will be used to convert Marshmallow schemas
        to the corresponding types of swagger objects. These default to the
        global registries.

    :param marshmallow.Schema default_response_schema: Schema to use as the default of all responses
    """

    _open_api_version: str

    def __init__(
        self,
        openapi_major_version: int,
        version: str = "1.0.0",
        title: str = "My API",
        description: str = "",
        query_string_converter_registry: Optional[ConverterRegistry] = None,
        request_body_converter_registry: Optional[ConverterRegistry] = None,
        headers_converter_registry: Optional[ConverterRegistry] = None,
        response_converter_registry: Optional[ConverterRegistry] = None,
        default_response_schema: Schema = Error(),
        authenticator_converter_registry: Optional[
            AuthenticatorConverterRegistry
        ] = None,
        include_hidden: bool = False,
    ):
        self.include_hidden = include_hidden
        self.title = title
        self.version = version
        self.description = description
        self._query_string_converter = self._create_converter(
            query_string_converter_registry,
            global_query_string_converter_registry,
            openapi_major_version,
        )
        self._request_body_converter = self._create_converter(
            request_body_converter_registry,
            global_request_body_converter_registry,
            openapi_major_version,
        )
        self._headers_converter = self._create_converter(
            headers_converter_registry,
            global_headers_converter_registry,
            openapi_major_version,
        )
        self._response_converter = self._create_converter(
            response_converter_registry,
            global_response_converter_registry,
            openapi_major_version,
        )

        self.flask_converters_to_swagger_types = {
            "uuid": sw.string,
            "uuid_string": sw.string,
            "string": sw.string,
            "path": sw.string,
            "int": sw.integer,
            "float": sw.number,
        }

        self.authenticator_converter = self._create_authenticator_converter(
            authenticator_converter_registry,
            global_authenticator_converter_registry,
            openapi_major_version,
        )

        self.default_response_schema = default_response_schema

    def _get_info(self) -> Dict[str, str]:
        return {
            sw.version: self.version,
            sw.title: self.title,
            sw.description: self.description,
        }

    def _create_converter(
        self,
        converter_registry: Optional[ConverterRegistry],
        default_registry: ConverterRegistry,
        openapi_major_version: int,
    ) -> Callable:
        return functools.partial(
            (converter_registry or default_registry).convert,
            openapi_version=openapi_major_version,
        )

    def _create_authenticator_converter(
        self,
        converter_registry: Optional[AuthenticatorConverterRegistry],
        default_registry: AuthenticatorConverterRegistry,
        openapi_major_version: int,
    ) -> AuthenticatorConverter:
        registry: Any = type("authenticator_converter_registry", (), {})
        registry.get_security_schemes = functools.partial(
            (converter_registry or default_registry).get_security_schemes,
            openapi_version=openapi_major_version,
        )
        registry.get_security_requirements = functools.partial(
            (converter_registry or default_registry).get_security_requirements,
            openapi_version=openapi_major_version,
        )
        return registry

    def get_open_api_version(self) -> str:
        return self._open_api_version

    def register_flask_converter_to_swagger_type(
        self, flask_converter: str, swagger_type: Any
    ) -> None:
        """
        Register a converter for a type in flask to a swagger type.

        This can be used to alter the types of objects that already exist or add
        swagger types to objects that are added to the base flask configuration.

        For example, when adding custom path types to the Flask url_map, a
        converter can be added for customizing the swagger type.

        .. code-block:: python

            import enum

            from flask_rebar.swagger_generation import swagger_words as sw


            class TodoType(str, enum.Enum):
                user = "user"
                group = "group"


            class TodoTypeConverter:

                @staticmethod
                def to_swagger():
                    return {
                        sw.type_: sw.string,
                        sw.enum: [t.value for t in TodoType],
                    }

            @registry.handles(
                rule="/todos/<todo_type:type_>",
            )
            def get_todos_by_type(type_):
                ...

            generator.register_flask_converter_to_swagger_type(
                flask_converter="todo_type",
                swagger_type=TodoTypeConverter,
            )

        With the above example, when something is labeled as a ``todo_type`` in
        a path. The correct swagger can be returned.
        """
        self.flask_converters_to_swagger_types[flask_converter] = swagger_type



"""
    Test Rebar
    ~~~~~~~~~~

    Tests for the basic usage of Flask-Rebar.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import json
import unittest

import marshmallow as m
import marshmallow_objects as mo
from flask import Flask, make_response

from parametrize import parametrize

from flask_rebar import messages
from flask_rebar import HeaderApiKeyAuthenticator, SwaggerV3Generator
from flask_rebar.compat import set_data_key
from flask_rebar.rebar import Rebar
from flask_rebar.rebar import prefix_url
from flask_rebar.testing import validate_swagger
from flask_rebar.utils.defaults import USE_DEFAULT
from flask_rebar.testing.swagger_jsonschema import SWAGGER_V3_JSONSCHEMA

DEFAULT_AUTH_HEADER = "x-default-auth"
DEFAULT_AUTH_SECRET = "SECRET!"
DEFAULT_ALTERNATIVE_AUTH_HEADER = "x-api-default-auth"
DEFAULT_ALTERNATIVE_AUTH_SECRET = "ALSO A SECRET!"
DEFAULT_RESPONSE = {"uid": "0", "name": "I'm the default for testing!"}
DEFAULT_ERROR = (
    messages.internal_server_error._asdict()  # noqa - _asdict is NOT internal.
)


class FooSchema(m.Schema):
    uid = m.fields.String()
    name = m.fields.String()


class FooModel(mo.Model):
    uid = mo.fields.String()
    name = mo.fields.String()


class ListOfFooSchema(m.Schema):
    data = m.fields.Nested(FooSchema, many=True)


class ListOfFooModel(mo.Model):
    data = mo.NestedModel(FooModel, many=True)


class FooUpdateSchema(m.Schema):
    name = m.fields.String()


class FooUpdateModel(mo.Model):
    name = mo.fields.String()


class FooListSchema(m.Schema):
    name = m.fields.String(required=True)


class FooListModel(mo.Model):
    name = mo.fields.String(required=True)


class HeadersSchema(m.Schema):
    name = set_data_key(field=m.fields.String(required=True), key="x-name")


class HeadersModel(mo.Model):
    name = set_data_key(field=mo.fields.String(required=True), key="x-name")


class MeSchema(m.Schema):
    user_name = m.fields.String()


class DefaultResponseSchema(
    m.Schema
):  # DEFAULT_RESPONSE = {"uid": "0", "name": "I'm the default for testing!"}
    uid = m.fields.String()
    name = m.fields.String()


def get_swagger(test_client, prefix=None):
    url = "/swagger"
    if prefix:
        url = prefix_url(prefix=prefix, url=url)
    return test_client.get(url).json


def auth_headers(header=DEFAULT_AUTH_HEADER, secret=DEFAULT_AUTH_SECRET):
    return dict([(header, secret)])


def alternative_auth_headers(
    header=DEFAULT_ALTERNATIVE_AUTH_HEADER, secret=DEFAULT_ALTERNATIVE_AUTH_SECRET
):
    return dict([(header, secret)])


def create_rebar_app(rebar):
    app = Flask("RebarTest")
    app.testing = True
    rebar.init_app(app)
    return app


def register_default_authenticator(registry):
    default_authenticator = HeaderApiKeyAuthenticator(
        header=DEFAULT_AUTH_HEADER, name="default"
    )
    default_authenticator.register_key(app_name="internal", key=DEFAULT_AUTH_SECRET)
    registry.set_default_authenticator(default_authenticator)


def register_multiple_authenticators(registry):
    default_authenticator = HeaderApiKeyAuthenticator(
        header=DEFAULT_AUTH_HEADER, name="default"
    )
    default_authenticator.register_key(app_name="internal", key=DEFAULT_AUTH_SECRET)
    alternative_default_authenticator = HeaderApiKeyAuthenticator(
        header=DEFAULT_ALTERNATIVE_AUTH_HEADER, name="alternative"
    )
    alternative_default_authenticator.register_key(
        app_name="internal", key=DEFAULT_ALTERNATIVE_AUTH_SECRET
    )
    registry.set_default_authenticators(
        (default_authenticator, alternative_default_authenticator)
    )


def register_endpoint(
    registry,
    func=None,
    path="/foos/<foo_uid>",
    method="GET",
    endpoint=None,
    response_body_schema=None,
    query_string_schema=None,
    request_body_schema=None,
    headers_schema=None,
    authenticators=USE_DEFAULT,
):
    def default_handler_func(*args, **kwargs):
        return DEFAULT_RESPONSE

    registry.add_handler(
        func=func or default_handler_func,
        rule=path,
        method=method,
        endpoint=endpoint,
        response_body_schema=response_body_schema or {200: FooSchema()},
        query_string_schema=query_string_schema,
        request_body_schema=request_body_schema,
        headers_schema=headers_schema,
        authenticators=authenticators,
    )


class RebarTest(unittest.TestCase):
    def test_default_authentication(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        register_default_authenticator(registry)
        register_endpoint(registry)
        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/foos/1", headers=auth_headers())
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        resp = app.test_client().get(
            path="/foos/1", headers=auth_headers(secret="LIES!")
        )
        self.assertEqual(resp.status_code, 401)

    def test_default_authentication_w_multiple(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        register_multiple_authenticators(registry)
        register_endpoint(registry)
        app = create_rebar_app(rebar)

        # Test main
        resp = app.test_client().get(path="/foos/1", headers=auth_headers())
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        # Test alternative
        resp = app.test_client().get(path="/foos/1", headers=alternative_auth_headers())
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        resp = app.test_client().get(
            path="/foos/1", headers=auth_headers(secret="LIES!")
        )
        self.assertEqual(resp.status_code, 401)

    def test_override_authenticator(self):
        auth_header = "x-overridden-auth"
        auth_secret = "BLAM!"

        rebar = Rebar()
        registry = rebar.create_handler_registry()

        register_default_authenticator(registry)
        authenticator = HeaderApiKeyAuthenticator(header=auth_header)
        authenticator.register_key(app_name="internal", key=auth_secret)

        register_endpoint(registry, authenticators=authenticator)
        app = create_rebar_app(rebar)

        resp = app.test_client().get(
            path="/foos/1", headers=auth_headers(header=auth_header, secret=auth_secret)
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        # The default authentication doesn't work anymore!
        resp = app.test_client().get(path="/foos/1", headers=auth_headers())
        self.assertEqual(resp.status_code, 401)

    def test_override_with_no_authenticator(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        register_default_authenticator(registry)
        register_endpoint(registry, authenticators=None)
        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/foos/1")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

    @parametrize(
        "foo_cls,foo_update_cls,use_model",
        [(FooSchema, FooUpdateSchema, False), (FooModel, FooUpdateModel, True)],
    )
    def test_validate_body_parameters(self, foo_cls, foo_update_cls, use_model):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles(
            rule="/foos/<foo_uid>",
            method="PATCH",
            response_body_schema={200: foo_cls()},
            request_body_schema=foo_update_cls(),
        )
        def update_foo(foo_uid):
            if use_model:
                # Here we can also verify that in handler we get our expected FooModel type object
                self.assertIsInstance(rebar.validated_body, foo_update_cls)
                return foo_cls(name=rebar.validated_body.name, uid=foo_uid)
            else:
                return {"uid": foo_uid, "name": rebar.validated_body["name"]}

        app = create_rebar_app(rebar)

        resp = app.test_client().patch(
            path="/foos/1",
            data=json.dumps({"name": "jill"}),
            headers={"Content-Type": "application/json"},
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"uid": "1", "name": "jill"})

        resp = app.test_client().patch(
            path="/foos/1",
            data=json.dumps({"name": 123}),  # Name should be string, not int
            headers={"Content-Type": "application/json"},
        )
        self.assertEqual(resp.status_code, 400)

    @parametrize("foo_update_cls", [(FooUpdateSchema,), (FooUpdateModel,)])
    def test_flask_response_instance_interop_body_matches_schema(self, foo_update_cls):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        schema = foo_update_cls()

        @registry.handles(rule="/foo", method="PUT", response_body_schema=schema)
        def foo():
            return make_response((json.dumps({"name": "foo"}), {"foo": "bar"}))

        app = create_rebar_app(rebar)
        resp = app.test_client().put(path="/foo")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.headers["foo"], "bar")

    @parametrize("foo_update_cls", [(FooUpdateSchema,), (FooUpdateModel,)])
    def test_flask_response_instance_interop_body_does_not_match_schema(
        self, foo_update_cls
    ):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        schema = foo_update_cls()

        @registry.handles(rule="/foo", method="PUT", response_body_schema=schema)
        def foo():
            return make_response(json.dumps({"does not match": "foo schema"}))

        app = create_rebar_app(rebar)
        resp = app.test_client().put(path="/foo")
        self.assertEqual(resp.status_code, 500)

    def test_flask_response_instance_interop_no_schema(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles(rule="/foo", response_body_schema={302: None})
        def foo():
            return make_response(("Redirecting", 302, {"Location": "http://foo.com"}))

        app = create_rebar_app(rebar)
        resp = app.test_client().get(path="/foo")
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(resp.headers["Location"], "http://foo.com")

    @parametrize("list_of_foo_cls", [(ListOfFooSchema,), (ListOfFooModel,)])
    def test_validate_query_parameters(self, list_of_foo_cls):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles(
            rule="/foos",
            method="GET",
            response_body_schema={200: list_of_foo_cls()},
            query_string_schema=FooListSchema(),
        )
        def list_foos():
            return {"data": [{"name": rebar.validated_args["name"], "uid": "1"}]}

        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/foos?name=jill")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"data": [{"uid": "1", "name": "jill"}]})

        resp = app.test_client().get(path="/foos?foo=bar")  # missing required parameter
        self.assertEqual(resp.status_code, 400)

    @parametrize(
        "headers_cls, use_model", [(HeadersSchema, False), (HeadersModel, True)]
    )
    def test_validate_headers(self, headers_cls, use_model):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        register_default_authenticator(registry)

        @registry.handles(
            rule="/me",
            method="GET",
            response_body_schema={200: MeSchema()},
            headers_schema=headers_cls,
        )
        def get_me():
            name = (
                rebar.validated_headers.name
                if use_model
                else rebar.validated_headers["name"]
            )
            return {"user_name": name}

        app = create_rebar_app(rebar)

        headers = auth_headers()
        headers["x-name"] = "hello"

        resp = app.test_client().get(path="/me", headers=headers)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"user_name": "hello"})

        resp = app.test_client().get(
            path="/me", headers=auth_headers()  # Missing the x-name header!
        )
        self.assertEqual(resp.status_code, 400)

    def test_default_mimetype_for_null_response_schema(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry(default_mimetype="content/type")

        @registry.handles(rule="/me", method="DELETE", response_body_schema={204: None})
        def delete_me():
            return None, 204

        app = create_rebar_app(rebar)
        resp = app.test_client().delete(path="/me")

        self.assertEqual(resp.status_code, 204)
        self.assertEqual(resp.data.decode("utf-8"), "")
        self.assertEqual(resp.headers["Content-Type"], "content/type")

    def test_default_mimetype_for_non_null_response_schema(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry(default_mimetype="content/type")

        @registry.handles(
            rule="/me", method="DELETE", response_body_schema={204: m.Schema()}
        )
        def delete_me():
            return {}, 204

        app = create_rebar_app(rebar)
        resp = app.test_client().delete(path="/me")

        self.assertEqual(resp.status_code, 204)
        self.assertEqual(resp.data.decode("utf-8"), "")
        self.assertEqual(resp.headers["Content-Type"], "content/type")

    def test_handler_mimetype_for_null_response_schema(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles(
            rule="/me",
            method="DELETE",
            response_body_schema={204: None},
            mimetype="content/type",
        )
        def delete_me():
            return None, 204

        app = create_rebar_app(rebar)
        resp = app.test_client().delete(path="/me")

        self.assertEqual(resp.status_code, 204)
        self.assertEqual(resp.data.decode("utf-8"), "")
        self.assertEqual(resp.headers["Content-Type"], "content/type")

    def test_handler_mimetype_for_non_null_response_schema(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles(
            rule="/me",
            method="DELETE",
            response_body_schema={204: m.Schema()},
            mimetype="content/type",
        )
        def delete_me():
            return {}, 204

        app = create_rebar_app(rebar)
        resp = app.test_client().delete(path="/me")

        self.assertEqual(resp.status_code, 204)
        self.assertEqual(resp.data.decode("utf-8"), "")
        self.assertEqual(resp.headers["Content-Type"], "content/type")

    def test_handler_mimetype_overrides_default_mimetype(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry(default_mimetype="default/type")

        @registry.handles(
            rule="/me",
            method="DELETE",
            response_body_schema={204: m.Schema()},
            mimetype="handler/type",
        )
        def delete_me():
            return {}, 204

        app = create_rebar_app(rebar)
        resp = app.test_client().delete(path="/me")

        self.assertEqual(resp.status_code, 204)
        self.assertEqual(resp.data.decode("utf-8"), "")
        self.assertEqual(resp.headers["Content-Type"], "handler/type")

    @parametrize("foo_definition", [(FooSchema,), (FooModel,)])
    def test_view_function_tuple_response(self, foo_definition):
        header_key = "X-Foo"
        header_value = "bar"
        headers = {header_key: header_value}

        for (
            response_body_schema,
            rv,
            expected_status,
            expected_body,
            expected_headers,
        ) in [
            (foo_definition(), DEFAULT_RESPONSE, 200, DEFAULT_RESPONSE, {}),
            (
                {201: foo_definition()},
                (DEFAULT_RESPONSE, 201),
                201,
                DEFAULT_RESPONSE,
                {},
            ),
            ({201: foo_definition()}, (DEFAULT_RESPONSE, 200), 500, DEFAULT_ERROR, {}),
            ({204: None}, (None, 204), 204, "", {}),
            (
                {200: foo_definition()},
                (DEFAULT_RESPONSE, headers),
                200,
                DEFAULT_RESPONSE,
                headers,
            ),
            (
                {201: foo_definition()},
                (DEFAULT_RESPONSE, 201, headers),
                201,
                DEFAULT_RESPONSE,
                headers,
            ),
            ({201: None}, (None, 201, headers), 201, "", headers),
            ({201: foo_definition()}, ({}, 201, headers), 201, {}, headers),
        ]:
            rebar = Rebar()
            registry = rebar.create_handler_registry()

            @registry.handles(rule="/foo", response_body_schema=response_body_schema)
            def foo():
                return rv

            app = create_rebar_app(rebar)

            resp = app.test_client().get("/foo")

            body = resp.json if resp.data else resp.data.decode()
            self.assertEqual(body, expected_body)
            self.assertEqual(resp.status_code, expected_status)

            for key, value in expected_headers.items():
                self.assertEqual(resp.headers[key], value)

    def test_swagger_endpoint_is_automatically_created(self):
        rebar = Rebar()
        rebar.create_handler_registry()
        app = create_rebar_app(rebar)

        resp = app.test_client().get("/swagger")

        self.assertEqual(resp.status_code, 200)

        validate_swagger(resp.json)

    def test_swagger_ui_endpoint_is_automatically_created(self):
        rebar = Rebar()
        rebar.create_handler_registry()
        app = create_rebar_app(rebar)

        resp = app.test_client().get("/swagger/ui/")
        self.assertEqual(resp.status_code, 200)

    def test_swagger_ui_without_trailing_slash(self):
        rebar = Rebar()
        rebar.create_handler_registry()
        app = create_rebar_app(rebar)

        resp = app.test_client().get("/swagger/ui")
        self.assertEqual(resp.status_code, 200)

    def test_swagger_can_be_set_to_v3(self):
        rebar = Rebar()
        rebar.create_handler_registry(swagger_generator=SwaggerV3Generator())
        app = create_rebar_app(rebar)

        resp = app.test_client().get("/swagger")
        self.assertEqual(resp.status_code, 200)
        validate_swagger(resp.json, SWAGGER_V3_JSONSCHEMA)

        resp = app.test_client().get("/swagger/ui/")
        self.assertEqual(resp.status_code, 200)

    @parametrize("foo_definition", [(FooSchema(),), (FooModel(),)])
    def test_register_multiple_paths(self, foo_definition):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        common_kwargs = {"method": "GET", "response_body_schema": {200: foo_definition}}

        @registry.handles(rule="/bars/<foo_uid>", endpoint="bar", **common_kwargs)
        @registry.handles(rule="/foos/<foo_uid>", endpoint="foo", **common_kwargs)
        def handler_func(foo_uid):
            return DEFAULT_RESPONSE

        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/foos/1")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        resp = app.test_client().get(path="/bars/1")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        swagger = get_swagger(test_client=app.test_client())
        self.assertIn("/bars/{foo_uid}", swagger["paths"])
        self.assertIn("/foos/{foo_uid}", swagger["paths"])

    @parametrize("foo_definition", [(FooSchema(),), (FooModel(),)])
    def test_register_multiple_methods(self, foo_definition):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        common_kwargs = {
            "rule": "/foos/<foo_uid>",
            "response_body_schema": {200: foo_definition},
        }

        @registry.handles(method="GET", endpoint="get_foo", **common_kwargs)
        @registry.handles(method="PATCH", endpoint="update_foo", **common_kwargs)
        def handler_func(foo_uid):
            return DEFAULT_RESPONSE

        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/foos/1")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        resp = app.test_client().patch(path="/foos/1")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, DEFAULT_RESPONSE)

        resp = app.test_client().post(path="/foos/1")
        self.assertEqual(resp.status_code, 405)

        swagger = get_swagger(test_client=app.test_client())
        self.assertIn("get", swagger["paths"]["/foos/{foo_uid}"])
        self.assertIn("patch", swagger["paths"]["/foos/{foo_uid}"])

    @parametrize(
        "headers_def, use_model", [(HeadersSchema(), False), (HeadersModel, True)]
    )
    def test_default_headers(self, headers_def, use_model):
        rebar = Rebar()
        registry = rebar.create_handler_registry()
        registry.set_default_headers_schema(headers_def)

        @registry.handles(rule="/me", method="GET", response_body_schema=MeSchema())
        def get_me():
            name = (
                rebar.validated_headers.name
                if use_model
                else rebar.validated_headers["name"]
            )
            return {"user_name": name}

        @registry.handles(
            rule="/myself",
            method="GET",
            response_body_schema=DefaultResponseSchema(),
            # Let's make sure this can be overridden
            headers_schema=None,
        )
        def get_myself():
            return DEFAULT_RESPONSE

        app = create_rebar_app(rebar)

        resp = app.test_client().get(path="/me", headers={"x-name": "hello"})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"user_name": "hello"})

        resp = app.test_client().get(path="/me")
        self.assertEqual(resp.status_code, 400)

        resp = app.test_client().get(path="/myself")
        self.assertEqual(resp.status_code, 200)

        swagger = get_swagger(test_client=app.test_client())
        self.assertEqual(
            swagger["paths"]["/me"]["get"]["parameters"][0]["name"], "x-name"
        )
        self.assertNotIn("parameters", swagger["paths"]["/myself"]["get"])

    def test_swagger_endpoints_can_be_omitted(self):
        rebar = Rebar()
        rebar.create_handler_registry(spec_path=None, swagger_ui_path=None)
        app = create_rebar_app(rebar)

        resp = app.test_client().get("/swagger")
        self.assertEqual(resp.status_code, 404)

        resp = app.test_client().get("/swagger/ui")
        self.assertEqual(resp.status_code, 404)

    def test_rebar_can_be_url_prefixed(self):
        app = Flask(__name__)
        app.testing = True

        rebar = Rebar()
        registry_v1 = rebar.create_handler_registry(prefix="v1")
        registry_v2 = rebar.create_handler_registry(
            prefix="/v2/"
        )  # Slashes shouldn't matter

        # We use the same endpoint to show that the swagger operationId gets set correctly
        # and the Flask endpoint gets prefixed
        register_endpoint(registry=registry_v1, endpoint="get_foo")
        register_endpoint(registry=registry_v2, endpoint="get_foo")

        rebar.init_app(app)

        for prefix in ("v1", "v2"):
            resp = app.test_client().get(prefix_url(prefix=prefix, url="/swagger/ui/"))
            self.assertEqual(resp.status_code, 200)

            swagger = get_swagger(test_client=app.test_client(), prefix=prefix)
            validate_swagger(swagger)

            self.assertEqual(
                swagger["paths"][prefix_url(prefix=prefix, url="/foos/{foo_uid}")][
                    "get"
                ]["operationId"],
                "get_foo",
            )
            resp = app.test_client().get(prefix_url(prefix=prefix, url="/foos/1"))
            self.assertEqual(resp.status_code, 200)

    def test_clone_rebar(self):
        rebar = Rebar()
        app = Flask(__name__)
        app.testing = True

        registry = rebar.create_handler_registry()

        register_default_authenticator(registry)
        register_endpoint(registry)

        cloned = registry.clone()
        cloned.prefix = "v1"
        rebar.add_handler_registry(registry=cloned)

        rebar.init_app(app)

        resp = app.test_client().get("/swagger/ui/")
        self.assertEqual(resp.status_code, 200)

        resp = app.test_client().get(path="/foos/1", headers=auth_headers())
        self.assertEqual(resp.status_code, 200)

        resp = app.test_client().get("/v1/swagger/ui/")
        self.assertEqual(resp.status_code, 200)

        resp = app.test_client().get(path="/v1/foos/1", headers=auth_headers())
        self.assertEqual(resp.status_code, 200)

    def test_uncaught_errors_are_not_jsonified_in_debug_mode(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles("/force_500")
        def force_500():
            raise ArithmeticError()

        app = create_rebar_app(rebar)

        app.debug = False
        resp = app.test_client().get(path="/force_500")
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.content_type, "application/json")

        app.debug = True
        with self.assertRaises(ArithmeticError):
            app.test_client().get(path="/force_500")

    def test_redirects_for_missing_trailing_slash(self):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        register_endpoint(registry=registry, path="/with_trailing_slash/")

        app = create_rebar_app(rebar)

        app.debug = False
        resp = app.test_client().get(path="/with_trailing_slash")
        self.assertIn(resp.status_code, (301, 308))
        self.assertTrue(resp.headers["Location"].endswith("/with_trailing_slash/"))

        app.debug = True
        resp = app.test_client().get(path="/with_trailing_slash")
        self.assertIn(resp.status_code, (301, 308))
        self.assertTrue(resp.headers["Location"].endswith("/with_trailing_slash/"))

    @parametrize(
        "foo_cls, foo_list_cls, headers_cls",
        [
            (FooSchema, FooListSchema, HeadersSchema),
            (FooModel, FooListModel, HeadersModel),
        ],
    )
    def test_bare_class_schemas_handled(self, foo_cls, foo_list_cls, headers_cls):
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        expected_headers = {"x-name": "Header Name"}

        def get_foo(*args, **kwargs):
            return expected_foo

        def post_foo(*args, **kwargs):
            return expected_foo

        register_endpoint(
            registry=registry,
            method="GET",
            path="/my_get_endpoint",
            headers_schema=headers_cls,
            response_body_schema={200: foo_cls},
            query_string_schema=foo_list_cls,
            func=get_foo,
        )

        register_endpoint(
            registry=registry,
            method="POST",
            path="/my_post_endpoint",
            request_body_schema=foo_list_cls,
            response_body_schema=foo_cls,
            func=post_foo,
        )

        expected_foo = FooSchema().load({"uid": "some_uid", "name": "Namey McNamerton"})

        app = create_rebar_app(rebar)
        # violate headers schema:
        resp = app.test_client().get(path="/my_get_endpoint?name=QuerystringName")
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(
            resp.json["message"], messages.header_validation_failed.message
        )

        # violate querystring schema:
        resp = app.test_client().get(path="/my_get_endpoint", headers=expected_headers)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(
            resp.json["message"], messages.query_string_validation_failed.message
        )
        # valid request:
        resp = app.test_client().get(
            path="/my_get_endpoint?name=QuerystringName", headers=expected_headers
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, expected_foo)

        resp = app.test_client().post(
            path="/my_post_endpoint",
            data='{"wrong": "Posted Name"}',
            content_type="application/json",
        )
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json["message"], messages.body_validation_failed.message)

        resp = app.test_client().post(
            path="/my_post_endpoint",
            data='{"name": "Posted Name"}',
            content_type="application/json",
        )
        self.assertEqual(resp.status_code, 200)

        # ensure Swagger generation doesn't break (Issue #115)
        from flask_rebar import SwaggerV2Generator, SwaggerV3Generator

        swagger = SwaggerV2Generator().generate(registry)
        self.assertIsNotNone(swagger)  # really only care that it didn't barf
        swagger = SwaggerV3Generator().generate(registry)
        self.assertIsNotNone(swagger)



from flask import Flask
from flask_rebar import HeaderApiKeyAuthenticator
from marshmallow import fields, pre_dump, pre_load

from flask_rebar import (
    Rebar,
    errors,
    HeaderApiKeyAuthenticator,
    Tag,
    SwaggerV2Generator,
)
from flask_rebar.validation import RequestSchema, ResponseSchema

from .converters import TodoTypeConverter


rebar = Rebar()

# Rebar will create a default swagger generator if none is specified.
# However, if you want more control over how the swagger is generated, you can
# provide your own.
# Here we've specified additional metadata for operation tags.
generator = SwaggerV2Generator(
    tags=[Tag(name="todo", description="Operations for managing TODO items.")]
)

registry = rebar.create_handler_registry(
    swagger_generator=generator,
    handlers="todo.handlers",
)


def create_app():
    app = Flask(__name__)

    # register new type for url mapping
    app.url_map.converters["todo_types"] = TodoTypeConverter
    generator.register_flask_converter_to_swagger_type("todo_types", TodoTypeConverter)

    authenticator = HeaderApiKeyAuthenticator(header="X-MyApp-Key")
    # The HeaderApiKeyAuthenticator does super simple authentication, designed for
    # service-to-service authentication inside of a protected network, by looking for a
    # shared secret in the specified header. Here we define what that shared secret is.
    authenticator.register_key(key="my-api-key")
    registry.set_default_authenticator(authenticator=authenticator)

    rebar.init_app(app=app)

    return app


if __name__ == "__main__":
    create_app().run()



"""
    Swagger Objects
    ~~~~~~~~~~~~~~~

    Python representations of select Swagger Objects.

    These are objects that aren't extractable from Flask or the Flask-Rebar handler registries
    and need to be manually set.

    :copyright: Copyright 2019 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from typing import Any, Dict, List, Optional, Union

from flask_rebar.swagger_generation import swagger_words as sw


class ExternalDocumentation:
    """Represents a Swagger "External Documentation Object"

    :param str url: The URL for the target documentation. Value MUST be in the format of a URL
    :param str description: A short description of the target documentation
    """

    def __init__(self, url: str, description: Optional[str] = None) -> None:
        self.url = url
        self.description = description

    def as_swagger(self) -> Dict[str, str]:
        """Create a Swagger representation of this object

        :rtype: dict
        """
        doc = {sw.url: self.url}
        if self.description:
            doc[sw.description] = self.description
        return doc


class Tag:
    """Represents a Swagger "Tag Object"

    :param str name: The name of the tag
    :param str description: A short description for the tag
    :param ExternalDocumentation external_docs: Additional external documentation for this tag
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        external_docs: Optional[ExternalDocumentation] = None,
    ) -> None:
        self.name = name
        self.description = description
        self.external_docs = external_docs

    def as_swagger(self) -> Dict[str, Union[str, Dict[str, str]]]:
        """Create a Swagger representation of this object

        :rtype: dict
        """
        doc: Dict[str, Union[str, Dict[str, str]]] = {sw.name: self.name}
        if self.description:
            doc[sw.description] = self.description
        if self.external_docs:
            doc[sw.external_docs] = self.external_docs.as_swagger()
        return doc


class ServerVariable:
    """Represents a Swagger "Server Variable Object"

    :param str default:
    :param str description:
    :param list[str] enum:
    """

    def __init__(
        self,
        default: str,
        description: Optional[str] = None,
        enum: Optional[List[str]] = None,
    ) -> None:
        self.default = default
        self.description = description
        self.enum = enum

    def as_swagger(self) -> Dict[str, Union[str, List[str]]]:
        """Create a Swagger representation of this object

        :rtype: dict
        """
        doc: Dict[str, Union[str, List[str]]] = {sw.default: self.default}
        if self.description:
            doc[sw.description] = self.description
        if self.enum:
            doc[sw.enum] = self.enum
        return doc


class Server:
    """Represents a Swagger "Server Object"

    :param str url:
    :param str description:
    :param dict[str, ServerVariable] variables:
    """

    def __init__(
        self,
        url: str,
        description: Optional[str] = None,
        variables: Optional[Dict[str, ServerVariable]] = None,
    ) -> None:
        self.url = url
        self.description = description
        self.variables = variables

    def as_swagger(self) -> Dict[str, Any]:
        """Create a Swagger representation of this object

        :rtype: dict
        """
        doc: Dict[str, Any] = {sw.url: self.url}
        if self.description:
            doc[sw.description] = self.description
        if self.variables:
            doc[sw.variables] = {k: v.as_swagger() for k, v in self.variables.items()}
        return doc



from collections import namedtuple
from typing import Type

from flask_rebar.authenticators import Authenticator, HeaderApiKeyAuthenticator
from .marshmallow_to_swagger import UnregisteredType
from . import swagger_words as sw
from typing import Any, Callable, Dict, Iterable, List, Optional


_Context = namedtuple(
    "_Context",
    [
        # The major version of OpenAPI being converter for
        "openapi_version"
    ],
)


class AuthenticatorConverter:
    """
    Abstract class for objects that convert Authenticator objects to
    security JSONSchema.

    When implementing your own AuthenticatorConverter you will need to:

    1) Set AUTHENTICATOR_TYPE to be your custom Authenticator class.

    2) Configure get_security_scheme to return a map of Security Scheme Objects.
        You should check for the openapi_version in context.openapi_version;
        If generation is requested for a version of the OpenAPI specification
        you do not intend to support,  we recommend raising a NotImplementedError.

    3) Configure get_security_requirements to return a list of Security Requirement Objects.
        You should check for the openapi_version in context.openapi_version;
        If generation is requested for a version of the OpenAPI specification
        you do not intend to support, we recommend raising a NotImplementedError.

    """

    AUTHENTICATOR_TYPE: Type[Authenticator]

    def get_security_schemes(
        self, obj: Authenticator, context: Optional[_Context] = None
    ) -> Dict[str, Any]:
        """
        Get the security schemes for the provided Authenticator object.

        OpenAPI Specification for defining security schemes
        2.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#security-definitions-object
        3.1.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.0.md#fixed-fields-6
                (see securitySchemes field)

        OpenAPI Specification for Security Scheme Object
        2.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#security-scheme-object
        3.1.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.0.md#security-scheme-object


        Example: An authenticator that makes use of an api_key and an application_key scheme
        {
            "api_key": {
                "type: "apiKey",
                "in": "header",
                "name": "X-API-Key"
            },
            "application_key" : {
                "type": "apiKey",
                "in": "query",
                "name": "application-key"
            }
        }

        Note: It is fine for multiple Authenticators to share Security Scheme definitions. Each Authenticator should
                return all scheme definitions that it makes use of.

        :param flask.authenticators.Authenticator obj: Authenticator instance to generate swagger for.
                    You can assume this is of type AUTHENTICATOR_TYPE
        :param _Context context: The context swagger is being generated for.
        :rtype: dict: Key should be the name for the scheme, the value should be a Security Scheme Object
        """
        raise NotImplementedError()

    def get_security_requirements(
        self, obj: Authenticator, context: Optional[_Context] = None
    ) -> List[Any]:
        """
        Get the security requirements for the provided Authenticator object

        OpenAPI Specification for Security Requirement Object
        2.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#security-requirement-object
        3.1.0: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.0.md#security-requirement-object

        Example: Require oauth with scope "read:stuff" OR api_key AND application_key
        [
            {"oauth": ["read:stuff"] },
            {"api_key": [], "application_key": []}
        ]

        :param flask_rebar.authenticators.Authenticator obj:
        :param _Context context:
        :rtype: list
        """
        raise NotImplementedError()


def make_class_from_method(
    authenticator_class: Type[Authenticator], func: Callable
) -> Type[AuthenticatorConverter]:
    """
    Utility to handle converting old-style method converters into new-style AuthenticatorConverters.
    """
    name = authenticator_class.__name__ + "Converter"
    meta = {
        "AUTHENTICATOR_TYPE": authenticator_class,
        "get_security_schemes": lambda self, obj, context: dict([func(obj)]),
        "get_security_requirements": lambda self, obj, context: [{func(obj)[0]: []}],
    }
    return type(name, (AuthenticatorConverter,), meta)


class HeaderApiKeyConverter(AuthenticatorConverter):
    AUTHENTICATOR_TYPE = HeaderApiKeyAuthenticator

    def get_security_requirements(
        self, obj: Authenticator, context: Optional[_Context] = None
    ) -> List[Dict[str, List]]:
        """
        :param HeaderApiKeyAuthenticator obj:
        :param _Context context:
        :return: list
        """
        if not isinstance(obj, HeaderApiKeyAuthenticator):
            raise NotImplementedError("Only HeaderApiKeyAuthenticator is supported")
        return [{obj.name: []}]

    def get_security_schemes(
        self, obj: Authenticator, context: Optional[_Context] = None
    ) -> Dict[str, Dict[str, str]]:
        """
        :param HeaderApiKeyAuthenticator obj:
        :param _Context context:
        :return: dict
        """
        if not isinstance(obj, HeaderApiKeyAuthenticator):
            raise NotImplementedError("Only HeaderApiKeyAuthenticator is supported")
        return {
            obj.name: {sw.type_: sw.api_key, sw.in_: sw.header, sw.name: obj.header}
        }


class AuthenticatorConverterRegistry:
    def __init__(self) -> None:
        self._type_map: Dict[Type[Authenticator], AuthenticatorConverter] = {}

    def _convert(self, obj: Authenticator, context: _Context) -> None:
        pass

    def convert(self, obj: Authenticator, openapi_version: int = 2) -> Dict[str, Any]:
        raise RuntimeWarning("Use get_security_schemes or get_security_requirements")

    def register_type(self, converter: AuthenticatorConverter) -> None:
        """
        Registers a converter.

        :param AuthenticatorConverter converter:
        """
        self._type_map[converter.AUTHENTICATOR_TYPE] = converter

    def register_types(self, converters: Iterable[AuthenticatorConverter]) -> None:
        """
        Registers multiple converters.

        :param iterable[AuthenticatorConverter] converters:
        """
        for converter in converters:
            self.register_type(converter)

    def _get_converter_for_type(self, obj: Authenticator) -> AuthenticatorConverter:
        """
        Locates the registered converter for a given type.
        :param obj: instance to convert
        :return: converter for type of instance
        """
        method_resolution_order = obj.__class__.__mro__

        for cls in method_resolution_order:
            if cls in self._type_map:
                return self._type_map[cls]
        else:
            raise UnregisteredType(
                "No registered type found in method resolution order: {mro}\n"
                "Registered types: {types}".format(
                    mro=method_resolution_order, types=list(self._type_map.keys())
                )
            )

    def get_security_schemes(
        self, authenticator: Authenticator, openapi_version: int = 2
    ) -> Dict[str, Any]:
        """
        Get the security schemes for the provided Authenticator object

        :param flask.authenticators.Authenticator obj:
        :param int openapi_version: major version of OpenAPI to convert obj for
        :rtype: dict
        """
        # Remove this once legacy is gone
        if not isinstance(authenticator, Authenticator):
            return self.get_security_schemes_legacy(registry=authenticator)
        return self._get_converter_for_type(authenticator).get_security_schemes(
            authenticator, _Context(openapi_version=openapi_version)
        )

    def get_security_requirements(
        self, authenticator: Authenticator, openapi_version: int = 2
    ) -> List[Dict[str, Any]]:
        """
        Get the security requirements for the provided Authenticator object

        :param flask_rebar.authenticators.Authenticator obj:
        :param int openapi_version: major version of OpenAPI to convert obj for
        :rtype: list
        """
        return self._get_converter_for_type(authenticator).get_security_requirements(
            authenticator, _Context(openapi_version=openapi_version)
        )


authenticator_converter_registry = AuthenticatorConverterRegistry()
authenticator_converter_registry.register_types((HeaderApiKeyConverter(),))



"""
    Test Generic Utilities
    ~~~~~~~~~~~~~~~~~~~~~~

    Tests for the generic (i.e., non-request) utilities.

    :copyright: Copyright 2019 Autodesk, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import unittest
import warnings
from flask_rebar.utils.deprecation import deprecated, deprecated_parameters
from flask_rebar.utils.deprecation import config as deprecation_config


@deprecated_parameters(
    old_param1="new_param1",  # rename with no predicted end-of-life version
    old_param2=("new_param2", "v99"),  # rename with predicted end-of-life version
    old_param3=("new_param3",),  # rename with a poorly formed tuple
    old_param4=("new_param4", None),  # rename with explicitly None end-of-life version
    old_param5=(None, "v99.5"),  # no rename with explicit end-of-life version
    old_param6=None,  # deprecated param with no replacement, no specific end-of-life-version
    old_param7=(None, None),  # same as 6, but for the truly pedantic
    old_param8=(),  # could imagine someone accidentally doing this.. :P
)
def _add(
    new_param1=0,
    new_param2=0,
    new_param3=0,
    new_param4=0,
    old_param5=0,
    old_param6=0,
    old_param7=0,
    old_param8=0,
):
    return (
        new_param1
        + new_param2
        + new_param3
        + new_param4
        + old_param5
        + old_param6
        + old_param7
        + old_param8
    )


@deprecated()
def _deprecated_func1():
    return 1


@deprecated("new_func2")
def _deprecated_func2():
    return 2


@deprecated(("new_func3", "99"))
def _deprecated_func3():
    return 3


class TestParameterDeprecation(unittest.TestCase):
    def test_parameter_deprecation_none(self):
        """Function with deprecated params, called with new (or no) names used does not warn"""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # test with unnamed args
            result = _add(1, 2)
            self.assertEqual(result, 3)
            self.assertEqual(len(w), 0)
            # test with "new" named args
            result = _add(new_param1=3, new_param2=5)
            self.assertEqual(result, 8)
            self.assertEqual(len(w), 0)

    def test_parameter_deprecation_warnings(self):
        """Function with deprecated param names warns (with expiration version if specified)"""
        # without version spec
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(old_param1=1, new_param2=2)
            self.assertEqual(result, 3)
            self.assertEqual(len(w), 1)
            self.assertIn("old_param1", str(w[0].message))
            self.assertNotIn("new_param2", str(w[0].message))
            self.assertIs(w[0].category, FutureWarning)

        # with version spec
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(new_param1=1, old_param2=2)
            self.assertEqual(result, 3)
            self.assertEqual(len(w), 1)
            self.assertNotIn("old_param1", str(w[0].message))
            self.assertIn("new_param2", str(w[0].message))
            self.assertIn("v99", str(w[0].message))
            self.assertIs(w[0].category, FutureWarning)

        # with both
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(old_param1=1, old_param2=2)
            self.assertEqual(result, 3)
            self.assertEqual(len(w), 2)
            msg1 = str(w[0].message)
            msg2 = str(w[1].message)
            self.assertTrue(
                ("old_param1" in msg1 and "old_param2" in msg2)
                or ("old_param1" in msg2 and "old_param2" in msg1)
            )
            self.assertIs(w[0].category, FutureWarning)

        # with both (using poorly formed tuples)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(old_param3=3, old_param4=4)
            self.assertEqual(result, 7)
            self.assertEqual(len(w), 2)
            msg1 = str(w[0].message)
            msg2 = str(w[1].message)
            self.assertTrue(
                ("old_param3" in msg1 and "old_param4" in msg2)
                or ("old_param3" in msg2 and "old_param4" in msg1)
            )
            self.assertIs(w[0].category, FutureWarning)

        # with no replacement but specific expiration
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(new_param1=1, old_param5=5)
            self.assertEqual(result, 6)
            self.assertEqual(len(w), 1)
            msg = str(w[0].message)
            self.assertIn("old_param5 is deprecated", msg)
            self.assertIn("v99.5", msg)
            self.assertNotIn("new_param", msg)
            self.assertIs(w[0].category, FutureWarning)

        # with no replacement (specified as None)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(new_param1=1, old_param5=5)
            self.assertEqual(result, 6)
            self.assertEqual(len(w), 1)
            msg = str(w[0].message)
            self.assertIn("old_param5 is deprecated", msg)
            self.assertIn("v99.5", msg)
            self.assertNotIn("new_param", msg)
            self.assertIs(w[0].category, FutureWarning)

        # with no replacement -- specified as explicit (None, None) and implicit ()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(old_param7=7, old_param8=8)
            self.assertEqual(result, 15)
            self.assertEqual(len(w), 2)
            msgs = {str(w[0].message), str(w[1].message)}
            expected_msgs = {"old_param7 is deprecated", "old_param8 is deprecated"}
            self.assertEqual(expected_msgs, msgs)
            self.assertIn("v99.5", msg)
            self.assertNotIn("new_param", msg)

    def test_parameter_deprecation_warning_type(self):
        """Deprecation supports specifying type of warning"""
        deprecation_config.warning_type = DeprecationWarning
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _add(old_param1=50, new_param2=0)
            self.assertEqual(result, 50)
            self.assertEqual(len(w), 1)
            self.assertIs(w[0].category, DeprecationWarning)
        # reset (as deprecation_config is "global")
        deprecation_config.warning_type = FutureWarning


class TestFunctionDeprecation(unittest.TestCase):
    def test_bare_deprecation(self):
        """Deprecate function with no specified alternative"""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _deprecated_func1()
            self.assertEqual(result, 1)
            self.assertEqual(len(w), 1)
            self.assertEqual(str(w[0].message), "_deprecated_func1 is deprecated")

    def test_versionless_replacement(self):
        """Deprecate function with specified alternative, no end-of-life version"""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _deprecated_func2()
            self.assertEqual(result, 2)
            self.assertEqual(len(w), 1)
            self.assertIn("_deprecated_func2 is deprecated", str(w[0].message))
            self.assertIn("use new_func2", str(w[0].message))
            self.assertNotIn("version", str(w[0].message))

    def test_versioned_replacement(self):
        """Deprecate function with specified alternative, specified end-of-life version"""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = _deprecated_func3()
            self.assertEqual(result, 3)
            self.assertEqual(len(w), 1)
            self.assertIn("_deprecated_func3 is deprecated", str(w[0].message))
            self.assertIn("use new_func3", str(w[0].message))
            self.assertIn("version 99", str(w[0].message))



"""
    Swagger Words
    ~~~~~~~~~~~~~

    Python friendly aliases to reserved Swagger words.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from typing import Final

additional_properties: Final = "additionalProperties"
all_of: Final = "allOf"
allow_empty_value: Final = "allowEmptyValue"
any_of: Final = "anyOf"
api_key: Final = "apiKey"
array: Final = "array"
basic: Final = "basic"
binary: Final = "binary"
body: Final = "body"
boolean: Final = "boolean"
byte: Final = "byte"
collection_format: Final = "collectionFormat"
components: Final = "components"
consumes: Final = "consumes"
content: Final = "content"
csv: Final = "csv"
date: Final = "date"
date_time: Final = "date-time"
default: Final = "default"
definitions: Final = "definitions"
description: Final = "description"
double: Final = "double"
enum: Final = "enum"
external_docs: Final = "externalDocs"
exclusive_maximum: Final = "exclusiveMaximum"
exclusive_minimum: Final = "exclusiveMinimum"
explode: Final = "explode"
float_: Final = "float"
form: Final = "form"
format_: Final = "format"
header: Final = "header"
host: Final = "host"
in_: Final = "in"
info: Final = "info"
integer: Final = "integer"
int32: Final = "int32"
int64: Final = "int64"
items: Final = "items"
max_items: Final = "maxItems"
max_length: Final = "maxLength"
max_properties: Final = "maxProperties"
maximum: Final = "maximum"
min_items: Final = "minItems"
min_length: Final = "minLength"
min_properties: Final = "minProperties"
minimum: Final = "minimum"
multi: Final = "multi"
multiple_of: Final = "multipleOf"
name: Final = "name"
null: Final = "null"
nullable: Final = "x-nullable"
number: Final = "number"
oauth2: Final = "oauth2"
object_: Final = "object"
one_of: Final = "oneOf"
openapi: Final = "openapi"
operation_id: Final = "operationId"
parameters: Final = "parameters"
password: Final = "password"
path: Final = "path"
paths: Final = "paths"
pattern: Final = "pattern"
produces: Final = "produces"
properties: Final = "properties"
query: Final = "query"
ref: Final = "$ref"
request_body: Final = "requestBody"
required: Final = "required"
responses: Final = "responses"
schema: Final = "schema"
schemas: Final = "schemas"
schemes: Final = "schemes"
security: Final = "security"
security_definitions: Final = "securityDefinitions"
security_schemes: Final = "securitySchemes"
servers: Final = "servers"
simple: Final = "simple"
string: Final = "string"
style: Final = "style"
summary: Final = "summary"
swagger: Final = "swagger"
tags: Final = "tags"
title: Final = "title"
type_: Final = "type"
unique_items: Final = "uniqueItems"
url: Final = "url"
uuid: Final = "uuid"
variables: Final = "variables"
version: Final = "version"



from flask_rebar.swagger_generation.swagger_objects import ExternalDocumentation
from flask_rebar.swagger_generation.swagger_generator_v2 import SwaggerV2Generator
from flask_rebar.swagger_generation.swagger_generator_v3 import SwaggerV3Generator
from flask_rebar.swagger_generation.swagger_objects import Tag
from flask_rebar.swagger_generation.swagger_objects import Server
from flask_rebar.swagger_generation.swagger_objects import ServerVariable
from flask_rebar.swagger_generation.marshmallow_to_swagger import sets_swagger_attr
from flask_rebar.swagger_generation.marshmallow_to_swagger import ConverterRegistry



# hack to reintroduce backward-compatibility as SwaggerV2Generator used to live in swagger_generator.py

from flask_rebar.swagger_generation.swagger_generator_base import SwaggerGenerator
from flask_rebar.swagger_generation.swagger_generator_v2 import (
    SwaggerV2Generator,
)  # noqa
from flask_rebar.swagger_generation.swagger_generator_v3 import (
    SwaggerV3Generator,
)  # noqa



"""
    Marshmallow to Swagger Conversion
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Utilities for converting Marshmallow objects to their
    corresponding Swagger representation.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import copy
import inspect
import logging
import sys
from collections import namedtuple
from typing import (
    overload,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

import marshmallow as m
from marshmallow import Schema
from marshmallow.validate import Range
from marshmallow.validate import OneOf
from marshmallow.validate import Length
from marshmallow.validate import Validator

from flask_rebar import compat
from flask_rebar.authenticators import Authenticator
from flask_rebar.validation import QueryParamList
from flask_rebar.validation import CommaSeparatedList
from flask_rebar.swagger_generation import swagger_words as sw

# for easier type hinting
MarshmallowObject = Union[Schema, m.fields.Field, Validator]
T = TypeVar("T", bound=MarshmallowObject)
TField = TypeVar("TField", bound=m.fields.Field)

# for type hinting decorators
S = TypeVar("S")
P = ParamSpec("P")

LoadDumpOptions = None
try:
    EnumField: Any = m.fields.Enum
except AttributeError:
    try:
        from marshmallow_enum import EnumField, LoadDumpOptions  # type: ignore
    except ImportError:
        EnumField = None


# Special value to signify that a JSONSchema field should be left unset
class UNSET:
    pass


# We'll use this to mark methods as JSONSchema attribute setters
_method_marker = "__sets_jsonschema_attr__"

# Holds attributes that we can pass around in these recursive
# calls to converters. Bit messy, but :shrug:
_Context = namedtuple(
    "_Context",
    [
        # This will hold a reference to a convert method that can be used
        # to make recursive calls
        "convert",
        # Only really using this for validators at the moment. It will hold the
        # JSONSchema object that's been converter so far, so that the validator
        # can be converted based on the type of the schema.
        "memo",
        # The current schema being converted.
        "schema",
        # The major version of OpenAPI being converter for
        "openapi_version",
    ],
)


class UnregisteredType(Exception):
    pass


@overload
def _normalize_validate(validate: None) -> None:
    ...


@overload
def _normalize_validate(
    validate: Union[Callable, Iterable[Callable]]
) -> Iterable[Callable]:
    ...


def _normalize_validate(
    validate: Optional[Union[Callable, Iterable[Callable]]]
) -> Optional[Iterable[Callable]]:
    """
    Coerces the validate attribute on a Marshmallow field to a consistent type.

    The validate attribute on a Marshmallow field can either be a single
    Validator or a collection of Validators.

    :param Validator|list[Validator] validate:
    :rtype: list[Validator]
    """
    if callable(validate):
        return [validate]
    else:
        return validate


def get_swagger_title(obj: MarshmallowObject) -> str:
    """
    Gets a title for the given object. This title will be used
    as a name/key for the object in swagger.

    :param obj:
    :rtype: str
    """
    if hasattr(obj, "__swagger_title__"):
        return obj.__swagger_title__
    elif hasattr(obj, "__name__"):
        return obj.__name__
    else:
        return obj.__class__.__name__


def sets_swagger_attr(attr: str) -> Callable:
    """
    Decorates a `MarshmallowConverter` method, marking it as an JSONSchema
    attribute setter.

    Example usage::

        class Converter(MarshmallowConverter):
            MARSHMALLOW_TYPE = String()

            @sets_swagger_attr('type')
            def get_type(obj, context):
                return 'string'

    This converter receives instances of `String` and converts it to a
    JSONSchema object that looks like `{'type': 'string'}`.

    :param str attr: The attribute to set
    """

    def wrapper(f: Callable[P, T]) -> Callable[P, T]:
        setattr(f, _method_marker, attr)
        return f

    return wrapper


def get_schema_fields(schema: Schema) -> List[Tuple[str, m.fields.Field]]:
    """Retrieve all the names and field objects for a marshmallow Schema

    :param m.Schema schema:
    :returns: Yields tuples of the field name and the field itself
    :rtype: typing.Iterator[typing.Tuple[str, m.fields.Field]]
    """
    fields: List[Tuple[str, m.fields.Field]] = []
    for name, field in schema.fields.items():
        prop = compat.get_data_key(field)
        fields.append((prop, field))
    return sorted(fields)


class MarshmallowConverter(Generic[T]):
    """
    Abstract class for objects that convert Marshmallow objects to
    JSONSchema dictionaries.
    """

    MARSHMALLOW_TYPE: Any = None

    def convert(self, obj: T, context: _Context) -> Dict[str, Union[str, bool]]:
        """
        Converts a Marshmallow object to a JSONSchema dictionary.

        This inspects the converter instance for methods that have been
        marked as attribute setters, calling them to set attributes on the
        resulting JSONSchema dictionary.

        :param m.Schema|m.fields.Field|Validator obj:
            The Marshmallow object to be converted
        :param _Context context:
            Various information to help the converter understand how to
            convert the object.
        :rtype: dict
        """
        jsonschema_obj = {}

        for _, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(method, _method_marker):
                val = method(obj, context)
                if val is not UNSET:
                    jsonschema_obj[getattr(method, _method_marker)] = val

        return jsonschema_obj


class SchemaConverter(MarshmallowConverter[Schema]):
    """Converts Marshmallow Schema objects."""

    MARSHMALLOW_TYPE = m.Schema

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: Schema, context: _Context) -> str:
        if obj.many:
            return sw.array
        else:
            return sw.object_

    @sets_swagger_attr(sw.items)
    def get_items(
        self, obj: Schema, context: _Context
    ) -> Union[Type[UNSET], MarshmallowObject]:
        if not obj.many:
            return UNSET

        singular_obj = copy.deepcopy(obj)
        singular_obj.many = False

        return context.convert(singular_obj, context)

    @sets_swagger_attr(sw.properties)
    def get_properties(
        self, obj: Schema, context: _Context
    ) -> Union[Type[UNSET], Dict[str, Any]]:
        if obj.many:
            return UNSET

        properties = {}

        for prop, field in get_schema_fields(obj):
            properties[prop] = context.convert(field, context)

        return properties

    @sets_swagger_attr(sw.required)
    def get_required(
        self, obj: Schema, context: _Context
    ) -> Union[Type[UNSET], List[str]]:
        if obj.many or obj.partial is True:
            return UNSET

        required: List[str] = []
        obj_partial_is_collection = m.utils.is_collection(obj.partial)

        for name, field in obj.fields.items():
            if field.required:
                prop = compat.get_data_key(field)
                if obj_partial_is_collection and obj.partial and prop in obj.partial:
                    continue
                required.append(prop)

        if required and not obj.ordered:
            required = sorted(required)
        return required if required else UNSET

    @sets_swagger_attr(sw.description)
    def get_description(
        self, obj: Schema, context: _Context
    ) -> Union[Type[UNSET], str]:
        if obj.many:
            return UNSET
        elif obj.__doc__:
            return obj.__doc__
        else:
            return UNSET

    @sets_swagger_attr(sw.title)
    def get_title(self, obj: Schema, context: _Context) -> Union[Type[UNSET], str]:
        if not obj.many:
            return get_swagger_title(obj)
        else:
            return UNSET

    @sets_swagger_attr(sw.additional_properties)
    def get_additional_properties(self, obj: Schema, context: _Context) -> bool:
        if obj.unknown in (m.RAISE, m.EXCLUDE):
            return False
        elif obj.unknown is m.INCLUDE:
            return True
        else:
            raise ValueError(
                f"Unexpected Schema.unknown value {obj.unknown} for {obj} "
            )


class FieldConverter(MarshmallowConverter, Generic[TField]):
    """
    Base Converter for Marshmallow Field objects.

    This should be extended for specific Field types.
    """

    MARSHMALLOW_TYPE: Type[m.fields.Field] = m.fields.Field

    def convert(self, obj: TField, context: _Context) -> Dict[str, Union[str, bool]]:
        jsonschema_obj = super().convert(obj, context)

        if obj.dump_only:
            jsonschema_obj["readOnly"] = True

        if obj.validate:
            validators = _normalize_validate(obj.validate)
            for validator in validators:
                try:
                    jsonschema_obj.update(
                        context.convert(
                            obj=validator,
                            context=_Context(
                                convert=context.convert,
                                memo=jsonschema_obj,
                                schema=context.schema,
                                openapi_version=context.openapi_version,
                            ),
                        )
                    )
                except UnregisteredType as e:
                    logging.debug(
                        "Unable to convert validator {validator}: {err}".format(
                            validator=validator, err=e
                        )
                    )

        return jsonschema_obj

    # With OpenApi 3.1 nullable has been removed entirely
    # and allowing 'none' means we return an array of allowed types that includes sw.null
    def null_type_determination(
        self, obj: TField, context: _Context, sw_type: str
    ) -> Union[str, List[str]]:
        if context.openapi_version == 3 and obj.allow_none:
            return [sw_type, sw.null]
        else:
            return sw_type

    @sets_swagger_attr(sw.default)
    def get_default(self, obj: TField, context: _Context) -> Any:
        if (
            obj.load_default is not m.missing
            # Marshmallow accepts a callable for the default. This is tricky
            # to handle, so let's just ignore this for now.
            and not callable(obj.load_default)
        ):
            return obj.load_default
        else:
            return UNSET

    @sets_swagger_attr(sw.nullable)
    def get_nullable(self, obj: TField, context: _Context) -> Union[Type[UNSET], bool]:
        if context.openapi_version == 2 and obj.allow_none is not False:
            return True
        else:
            return UNSET

    @sets_swagger_attr(sw.description)
    def get_description(
        self, obj: TField, context: _Context
    ) -> Union[Type[UNSET], str]:
        if "description" in obj.metadata:
            return obj.metadata["description"]
        else:
            return UNSET


class ValidatorConverter(MarshmallowConverter[Validator]):
    """
    Base Converter for Marshmallow Validator objects.

    This should be extended for specific Validator types.
    """

    MARSHMALLOW_TYPE: Union[Type[Validator], Type[OneOf]] = Validator


class NestedConverter(FieldConverter[m.fields.Nested]):
    MARSHMALLOW_TYPE = m.fields.Nested

    def convert(self, obj: m.fields.Nested, context: _Context) -> Dict[str, Any]:
        return context.convert(obj.schema, context)


class ListConverter(FieldConverter[m.fields.List]):
    MARSHMALLOW_TYPE = m.fields.List

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: m.fields.List, context: _Context) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.array)

    @sets_swagger_attr(sw.items)
    def get_items(
        self, obj: m.fields.List, context: _Context
    ) -> Union[Type[UNSET], m.fields.List]:
        return context.convert(obj.inner, context)


class DictConverter(FieldConverter[m.fields.Dict]):
    MARSHMALLOW_TYPE = m.fields.Dict

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: m.fields.Dict, context: _Context) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.object_)


class IntegerConverter(FieldConverter[m.fields.Integer]):
    MARSHMALLOW_TYPE = m.fields.Integer

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.Integer, context: _Context
    ) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.integer)


class StringConverter(FieldConverter[m.fields.String]):
    MARSHMALLOW_TYPE = m.fields.String

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.String, context: _Context
    ) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.string)


class NumberConverter(FieldConverter[m.fields.Number]):
    MARSHMALLOW_TYPE = m.fields.Number

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.Number, context: _Context
    ) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.number)


class BooleanConverter(FieldConverter[m.fields.Boolean]):
    MARSHMALLOW_TYPE = m.fields.Boolean

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.Boolean, context: _Context
    ) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.boolean)


class DateTimeConverter(FieldConverter[m.fields.DateTime]):
    MARSHMALLOW_TYPE = m.fields.DateTime

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.DateTime, context: _Context
    ) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.string)

    @sets_swagger_attr(sw.format_)
    def get_format(self, obj: m.fields.DateTime, context: _Context) -> str:
        return sw.date_time


class UUIDConverter(FieldConverter[m.fields.UUID]):
    MARSHMALLOW_TYPE = m.fields.UUID

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: m.fields.UUID, context: _Context) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.string)

    @sets_swagger_attr(sw.format_)
    def get_format(self, obj: m.fields.UUID, context: _Context) -> str:
        return sw.uuid


class DateConverter(FieldConverter[m.fields.Date]):
    MARSHMALLOW_TYPE = m.fields.Date

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: m.fields.Date, context: _Context) -> Union[str, List[str]]:
        return self.null_type_determination(obj, context, sw.string)

    @sets_swagger_attr(sw.format_)
    def get_format(self, obj: m.fields.Date, context: _Context) -> str:
        return sw.date


class MethodConverter(FieldConverter[m.fields.Method]):
    MARSHMALLOW_TYPE = m.fields.Method

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.Method, context: _Context
    ) -> Union[str, List[str]]:
        if "swagger_type" in obj.metadata:
            return self.null_type_determination(
                obj, context, obj.metadata["swagger_type"]
            )
        else:
            raise ValueError(
                'Must include "swagger_type" ' "keyword argument in Method field"
            )


class FunctionConverter(FieldConverter[m.fields.Function]):
    MARSHMALLOW_TYPE = m.fields.Function

    @sets_swagger_attr(sw.type_)
    def get_type(
        self, obj: m.fields.Function, context: _Context
    ) -> Union[str, List[str]]:
        if "swagger_type" in obj.metadata:
            return self.null_type_determination(
                obj, context, obj.metadata["swagger_type"]
            )
        else:
            raise ValueError(
                'Must include "swagger_type" ' "keyword argument in Function field"
            )


class ConstantConverter(FieldConverter[m.fields.Constant]):
    MARSHMALLOW_TYPE = m.fields.Constant

    @sets_swagger_attr(sw.enum)
    def get_enum(self, obj: m.fields.Constant, context: _Context) -> List[str]:
        return [obj.constant]


class CsvArrayConverter(ListConverter):
    MARSHMALLOW_TYPE = CommaSeparatedList

    @sets_swagger_attr(sw.collection_format)
    def get_collection_format(
        self, obj: CommaSeparatedList, context: _Context
    ) -> Union[Type[UNSET], str]:
        return sw.csv if context.openapi_version == 2 else UNSET

    @sets_swagger_attr(sw.style)
    def get_style(
        self, obj: CommaSeparatedList, context: _Context
    ) -> Union[Type[UNSET], str]:
        return sw.form if context.openapi_version == 3 else UNSET

    @sets_swagger_attr(sw.explode)
    def get_explode(
        self, obj: CommaSeparatedList, context: _Context
    ) -> Union[Type[UNSET], bool]:
        return False if context.openapi_version == 3 else UNSET


class MultiArrayConverter(ListConverter):
    MARSHMALLOW_TYPE = QueryParamList

    @sets_swagger_attr(sw.collection_format)
    def get_collection_format(
        self, obj: QueryParamList, context: _Context
    ) -> Union[Type[UNSET], str]:
        return sw.multi if context.openapi_version == 2 else UNSET

    @sets_swagger_attr(sw.explode)
    def get_explode(
        self, obj: QueryParamList, context: _Context
    ) -> Union[Type[UNSET], bool]:
        return True if context.openapi_version == 3 else UNSET


class RangeConverter(ValidatorConverter):
    MARSHMALLOW_TYPE = Range

    @sets_swagger_attr(sw.minimum)
    def get_minimum(
        self, obj: Range, context: _Context
    ) -> Union[Type[UNSET], int, float]:
        if obj.min is not None:
            return obj.min
        else:
            return UNSET

    @sets_swagger_attr(sw.maximum)
    def get_maximum(
        self, obj: Range, context: _Context
    ) -> Union[Type[UNSET], int, float]:
        if obj.max is not None:
            return obj.max
        else:
            return UNSET


class OneOfConverter(ValidatorConverter):
    MARSHMALLOW_TYPE = OneOf

    @sets_swagger_attr(sw.enum)
    def get_enum(self, obj: OneOf, context: _Context) -> List[str]:
        return list(obj.choices)


class LengthConverter(ValidatorConverter):
    MARSHMALLOW_TYPE = Length

    @sets_swagger_attr(sw.min_items)
    def get_minimum_items(
        self, obj: Length, context: _Context
    ) -> Union[Type[UNSET], int]:
        if context.memo[sw.type_] == sw.array:
            if obj.min is not None:
                return obj.min
        return UNSET

    @sets_swagger_attr(sw.max_items)
    def get_maximum_items(
        self, obj: Length, context: _Context
    ) -> Union[Type[UNSET], int]:
        if context.memo[sw.type_] == sw.array:
            if obj.max is not None:
                return obj.max
        return UNSET

    @sets_swagger_attr(sw.min_length)
    def get_minimum_length(
        self, obj: Length, context: _Context
    ) -> Union[Type[UNSET], int]:
        if context.memo[sw.type_] == sw.string:
            if obj.min is not None:
                return obj.min
        return UNSET

    @sets_swagger_attr(sw.max_length)
    def get_maximum_length(
        self, obj: Length, context: _Context
    ) -> Union[Type[UNSET], int]:
        if context.memo[sw.type_] == sw.string:
            if obj.max is not None:
                return obj.max
        return UNSET


class ConverterRegistry:
    """
    Registry for MarshmallowConverters.

    Schemas for responses, query strings, request bodies, etc. need to
    be converted differently. For example, response schemas as "dump"ed and
    request body schemas are "loaded". For another example, query strings
    don't support nesting.

    This registry also allows for additional converters to be added for custom
    Marshmallow types.
    """

    def __init__(self) -> None:
        self._type_map: Dict[
            Union[Type[MarshmallowObject], Type[Authenticator]], MarshmallowConverter
        ] = {}
        # self._validator_map = {}

    def register_type(self, converter: MarshmallowConverter) -> None:
        """
        Registers a converter.

        :param MarshmallowConverter converter:
        """
        self._type_map[converter.MARSHMALLOW_TYPE] = converter

    def register_types(self, converters: Iterable[MarshmallowConverter]) -> None:
        """
        Registers multiple converters.

        :param iterable[MarshmallowConverter] converters:
        """
        for converter in converters:
            self.register_type(converter)

    def _get_converter_for_type(self, obj: MarshmallowObject) -> MarshmallowConverter:
        """
        Locates the registered converter for a given type.
        :param obj: instance to convert
        :return: converter for type of instance
        """
        method_resolution_order = obj.__class__.__mro__

        for cls in method_resolution_order:
            if cls in self._type_map:
                return self._type_map[cls]
        else:
            raise UnregisteredType(
                "No registered type found in method resolution order: {mro}\n"
                "Registered types: {types}".format(
                    mro=method_resolution_order, types=list(self._type_map.keys())
                )
            )

    def _convert(
        self, obj: MarshmallowObject, context: _Context
    ) -> Dict[str, Union[str, bool]]:
        """
        Converts a Marshmallow object to a JSONSchema dictionary.

        :param m.Schema|m.fields.Field|Validator obj:
            The Marshmallow object to be converted
        :param _Context context:
            Various information to help the converter understand how to
            convert the given object.

            This helps with all the recursive nonsense.
        :rtype: dict
        """
        return self._get_converter_for_type(obj).convert(obj=obj, context=context)

    def convert(
        self, obj: MarshmallowObject, openapi_version: int = 2
    ) -> Dict[str, Any]:
        """
        Converts a Marshmallow object to a JSONSchema dictionary.

        :param m.Schema|m.fields.Field|Validator obj:
            The Marshmallow object to be converted
        :param int openapi_version: major version of OpenAPI to convert obj for
        :rtype: dict
        """
        return self._convert(
            obj=obj,
            context=_Context(
                convert=self._convert,
                memo={},
                schema=obj,
                openapi_version=openapi_version,
            ),
        )


class EnumConverter(FieldConverter):
    MARSHMALLOW_TYPE = EnumField  # type: ignore
    # Note that `obj` is typed as Any in this converter because mypy has great difficulty
    # trying to sort out type hints for EnumField, since it could be either m.fields.Enum
    # (marshmallow >= 3.18) or marshmallow_enum.EnumField

    @sets_swagger_attr(sw.type_)
    def get_type(self, obj: Any, context: _Context) -> Union[str, List[str]]:
        # Note: we don't (yet?) support mix-and-match between load_by and dump_by. Pick one.
        if obj.by_value or (
            LoadDumpOptions is not None
            and obj.load_by == obj.dump_by == LoadDumpOptions.value
        ):
            # I'm going out on a limb and assuming your enum uses same type for all vals, else caveat emptor:
            value_type = type(next(iter(obj.enum)).value)
            if value_type is int:
                return self.null_type_determination(obj, context, sw.integer)
            elif value_type is float:
                return self.null_type_determination(obj, context, sw.number)
            else:
                return self.null_type_determination(obj, context, sw.string)
        else:
            return self.null_type_determination(obj, context, sw.string)

    @sets_swagger_attr(sw.enum)
    def get_enum(self, obj: Any, context: _Context) -> List[str]:
        if obj.by_value or (
            LoadDumpOptions is not None
            and obj.load_by == obj.dump_by == LoadDumpOptions.value
        ):
            return [entry.value for entry in obj.enum]
        else:
            return [entry.name for entry in obj.enum]


ALL_CONVERTERS = tuple(
    [
        klass()
        for _, klass in inspect.getmembers(sys.modules[__name__], inspect.isclass)
        if issubclass(klass, MarshmallowConverter)
    ]
)


def _common_converters() -> List[MarshmallowConverter]:
    """Instantiates the converters we use in ALL of the registries below"""
    converters: List[MarshmallowConverter] = [
        BooleanConverter(),
        DateConverter(),
        DateTimeConverter(),
        FunctionConverter(),
        IntegerConverter(),
        LengthConverter(),
        ListConverter(),
        MethodConverter(),
        NumberConverter(),
        OneOfConverter(),
        RangeConverter(),
        SchemaConverter(),
        StringConverter(),
        UUIDConverter(),
        ConstantConverter(),
    ]
    if EnumConverter.MARSHMALLOW_TYPE is not None:  # type: ignore
        converters.append(EnumConverter())

    return converters


query_string_converter_registry: ConverterRegistry = ConverterRegistry()
query_string_converter_registry.register_types(
    _common_converters() + [CsvArrayConverter(), MultiArrayConverter()]
)

headers_converter_registry: ConverterRegistry = ConverterRegistry()
headers_converter_registry.register_types(
    _common_converters() + [CsvArrayConverter(), MultiArrayConverter()]
)

request_body_converter_registry: ConverterRegistry = ConverterRegistry()
request_body_converter_registry.register_types(
    _common_converters() + [DictConverter(), NestedConverter()]
)

response_converter_registry: ConverterRegistry = ConverterRegistry()
response_converter_registry.register_types(
    _common_converters() + [DictConverter(), NestedConverter()]
)



"""
    Errors
    ~~~~~~

    Exceptions that get transformed to HTTP error responses.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from flask_rebar.messages import ErrorMessage
from typing import Any, Dict, Optional, Union


class HttpJsonError(Exception):
    """
    Abstract base class for exceptions that will be cause and transformed
    into an appropriate HTTP error response with a JSON body.

    These can be raised at any time during the handling of a request,
    and the Rebar extension will handling catching it and transforming it.

    This class itself shouldn't be used. Instead, use one of the subclasses.

    :param str msg:
        A human readable message to be included in the JSON error response
    :param dict additional_data:
        Dictionary of additional keys and values to be set in the JSON body.
        Note that these keys and values are added to the root object of the
        response, not nested under "additional_data".
    """

    default_message: str
    http_status_code: int

    def __init__(
        self,
        msg: Optional[Union[str, ErrorMessage]] = None,
        additional_data: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.error_message = msg or self.default_message
        self.additional_data = additional_data
        super().__init__(self.error_message)


class BadRequest(HttpJsonError):
    http_status_code, default_message = 400, "Bad Request"


class Unauthorized(HttpJsonError):
    http_status_code, default_message = 401, "Unauthorized"


class PaymentRequired(HttpJsonError):
    http_status_code, default_message = 402, "Payment Required"


class Forbidden(HttpJsonError):
    http_status_code, default_message = 403, "Forbidden"


class NotFound(HttpJsonError):
    http_status_code, default_message = 404, "Not Found"


class MethodNotAllowed(HttpJsonError):
    http_status_code, default_message = 405, "Method Not Allowed"


class NotAcceptable(HttpJsonError):
    http_status_code, default_message = 406, "Not Acceptable"


class ProxyAuthenticationRequired(HttpJsonError):
    http_status_code, default_message = 407, "Proxy Authentication Required"


class RequestTimeout(HttpJsonError):
    http_status_code, default_message = 408, "Request Timeout"


class Conflict(HttpJsonError):
    http_status_code, default_message = 409, "Conflict"


class Gone(HttpJsonError):
    http_status_code, default_message = 410, "Gone"


class LengthRequired(HttpJsonError):
    http_status_code, default_message = 411, "Length Required"


class PreconditionFailed(HttpJsonError):
    http_status_code, default_message = 412, "Precondition Failed"


class RequestEntityTooLarge(HttpJsonError):
    http_status_code, default_message = 413, "Request Entity Too Large"


class RequestUriTooLong(HttpJsonError):
    http_status_code, default_message = 414, "Request URI Too Long"


class UnsupportedMediaType(HttpJsonError):
    http_status_code, default_message = 415, "Unsupported Media Type"


class RequestedRangeNotSatisfiable(HttpJsonError):
    http_status_code, default_message = 416, "Requested Range Not Satisfiable"


class ExpectationFailed(HttpJsonError):
    http_status_code, default_message = 417, "Expectation Failed"


class UnprocessableEntity(HttpJsonError):
    http_status_code, default_message = 422, "Unprocessable Entity"


class TooManyRequests(HttpJsonError):
    http_status_code, default_message = 429, "Too Many Requests"


class InternalError(HttpJsonError):
    http_status_code, default_message = 500, "Internal Server Error"


class NotImplemented(HttpJsonError):
    http_status_code, default_message = 501, "Not Implemented"


class BadGateway(HttpJsonError):
    http_status_code, default_message = 502, "Bad Gateway"


class ServiceUnavailable(HttpJsonError):
    http_status_code, default_message = 503, "Service Unavailable"


class GatewayTimeout(HttpJsonError):
    http_status_code, default_message = 504, "Gateway Timeout"



"""
    Test Errors
    ~~~~~~~~~~~

    Tests for the exception to HTTP error transformation.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import json
import unittest

from flask import Flask
from marshmallow import fields
from werkzeug.exceptions import BadRequest
from unittest.mock import ANY
from unittest.mock import patch
from tests.helpers import make_test_response

from flask_rebar import messages, validation, response, Rebar
from flask_rebar import errors
from flask_rebar.utils.request_utils import get_json_body_params_or_400
from flask_rebar.utils.request_utils import get_query_string_params_or_400


class TestErrors(unittest.TestCase):
    ERROR_MSG = messages.ErrorMessage(
        "Bamboozled!", messages.ErrorCode.INTERNAL_SERVER_ERROR
    )

    def setUp(self):
        self.app = self.create_app()
        self.app.response_class = make_test_response(self.app.response_class)

    def create_app(self):
        app = Flask(__name__)

        @app.route("/errors", methods=["GET"])
        def a_terrible_handler():
            raise errors.Conflict(msg=TestErrors.ERROR_MSG)

        @app.route("/uncaught_errors", methods=["GET"])
        def an_even_worse_handler():
            raise ArithmeticError()

        @app.route("/verbose_errors", methods=["GET"])
        def a_fancy_handler():
            raise errors.Conflict(
                msg=TestErrors.ERROR_MSG, additional_data={"foo": "bar"}
            )

        @app.route("/route_that_fails_validation", methods=["GET"])
        def validation_fails_handler():
            raise BadRequest()

        @app.route("/slow", methods=["GET"])
        def a_slow_handler():
            raise SystemExit()

        Rebar().init_app(app=app)

        return app

    def test_custom_http_errors_are_handled(self):
        resp = self.app.test_client().get("/errors")
        self.assertEqual(resp.status_code, 409)
        self.assertEqual(resp.content_type, "application/json")
        self.assertEqual(resp.json, TestErrors.ERROR_MSG._asdict())

    def test_custom_http_errors_can_have_additional_data(self):
        resp = self.app.test_client().get("/verbose_errors")
        self.assertEqual(resp.status_code, 409)
        self.assertEqual(resp.content_type, "application/json")
        expected = TestErrors.ERROR_MSG._asdict()
        expected["foo"] = "bar"
        self.assertEqual(resp.json, expected)

    def test_customize_rebar_error_attribute(self):
        rebar = self.app.extensions["rebar"]["instance"]
        # option 1: supply custom name for attribute in response
        rebar.error_code_attr = "xyz123"
        resp = self.app.test_client().get("/errors")
        expected = TestErrors.ERROR_MSG._asdict()
        expected["xyz123"] = expected.pop("rebar_error_code")
        self.assertEqual(resp.json, expected)

        # option 2: suppress rebar-internal error codes entirely
        rebar.error_code_attr = None
        resp = self.app.test_client().get("/errors")
        expected = TestErrors.ERROR_MSG._asdict()
        expected.pop("rebar_error_code")
        self.assertEqual(resp.json, expected)

    def test_default_400_errors_are_formatted_correctly(self):
        resp = self.app.test_client().get("/route_that_fails_validation")
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.content_type, "application/json")
        self.assertTrue(
            resp.json.get("message")
        )  # don't care about exact message wording, just existence

    def test_default_404_errors_are_formatted_correctly(self):
        resp = self.app.test_client().get("/nonexistent")
        self.assertEqual(resp.status_code, 404)
        self.assertEqual(resp.content_type, "application/json")
        self.assertTrue(
            resp.json.get("message")
        )  # don't care about exact message wording, just existence

    def test_default_405_errors_are_formatted_correctly(self):
        resp = self.app.test_client().put("/errors")
        self.assertEqual(resp.status_code, 405)
        self.assertEqual(resp.content_type, "application/json")
        self.assertTrue(
            resp.json.get("message")
        )  # don't care about exact message wording, just existence

    def test_default_500_errors_are_formatted_correctly(self):
        resp = self.app.test_client().get("/uncaught_errors")
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.content_type, "application/json")
        self.assertEqual(resp.json, messages.internal_server_error._asdict())

    def test_timeouts_log_exceptions(self):
        # in the wild, gunicorn or nginx will cutoff the wsgi server and return a 502
        # causing the wsgi server to raise SystemExit
        with patch.object(self.app.logger, "error") as mock_logger, self.assertRaises(
            SystemExit
        ):
            self.app.test_client().get("/slow")
            mock_logger.error.assert_called_with(
                "Exception on /slow [GET]", exc_info=ANY
            )


class TestJsonBodyValidation(unittest.TestCase):
    def setUp(self):
        self.app = self.create_app()
        self.app.response_class = make_test_response(self.app.response_class)

    def post_json(self, path, data):
        return self.app.test_client().post(
            path=path,
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
        )

    def create_app(self):
        app = Flask(__name__)
        Rebar().init_app(app=app)

        class NestedSchema(validation.RequestSchema):
            baz = fields.List(fields.Integer())

        class Schema(validation.RequestSchema):
            foo = fields.Integer(required=True)
            bar = fields.Email()
            nested = fields.Nested(NestedSchema)

        @app.route("/stuffs", methods=["POST"])
        def json_body_handler():
            data = get_json_body_params_or_400(schema=Schema)
            return response(data)

        return app

    def test_json_encoding_validation(self):
        resp = self.app.test_client().post(
            path="/stuffs", headers={"Content-Type": "application/json"}
        )
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, messages.empty_json_body._asdict())

        resp = self.app.test_client().post(
            path="/stuffs",
            data=json.dumps({"foo": 1}),
            headers={"Content-Type": "text/csv"},
        )
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, messages.unsupported_content_type._asdict())

        resp = self.app.test_client().post(
            path="/stuffs",
            data="not valid json",
            headers={"Content-Type": "application/json"},
        )
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, messages.invalid_json._asdict())

    def test_json_body_parameter_validation(self):
        # Only field errors
        resp = self.post_json(
            path="/stuffs", data={"foo": "one", "bar": "not-an-email"}
        )
        expected = messages.body_validation_failed._asdict()
        expected["errors"] = {
            "foo": "Not a valid integer.",
            "bar": "Not a valid email address.",
        }

        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, expected)

        # Only general errors
        resp = self.post_json(
            path="/stuffs", data={"foo": 1, "baz": "This is an unexpected field!"}
        )
        expected = messages.body_validation_failed._asdict()
        expected["errors"] = {"baz": "Unknown field."}

        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, expected)

        # Both field errors and general errors
        resp = self.post_json(
            path="/stuffs", data={"baz": "This is an unexpected field!"}
        )
        self.assertEqual(resp.status_code, 400)
        expected = messages.body_validation_failed._asdict()
        expected["errors"] = {
            "baz": "Unknown field.",
            "foo": "Missing data for required field.",
        }

        self.assertEqual(resp.json, expected)

        # Happy path
        resp = self.post_json(path="/stuffs", data={"foo": 1})
        self.assertEqual(resp.status_code, 200)

    def test_representing_complex_errors(self):
        resp = self.post_json(
            path="/stuffs",
            data={
                "bam": "wow!",
                "nested": {"baz": ["one", "two"], "unexpected": "surprise!"},
            },
        )
        expected = messages.body_validation_failed._asdict()
        expected["errors"] = {
            "bam": "Unknown field.",
            "foo": "Missing data for required field.",
            "nested": {
                "unexpected": "Unknown field.",
                "baz": {"0": "Not a valid integer.", "1": "Not a valid integer."},
            },
        }

        self.assertEqual(resp.status_code, 400)

        self.assertEqual(resp.json, expected)

    def test_invalid_json_error(self):
        resp = self.app.test_client().post(
            path="/stuffs",
            data='"Im technically valid JSON, but not an object"',
            headers={"Content-Type": "application/json"},
        )
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, messages.invalid_json._asdict())


class TestQueryStringValidation(unittest.TestCase):
    def setUp(self):
        self.app = self.create_app()
        self.app.response_class = make_test_response(self.app.response_class)

    def create_app(self):
        app = Flask(__name__)
        Rebar().init_app(app=app)

        class Schema(validation.RequestSchema):
            foo = fields.Integer(required=True)
            bar = fields.Boolean()
            baz = validation.CommaSeparatedList(fields.Integer())

        @app.route("/stuffs", methods=["GET"])
        def query_string_handler():
            params = get_query_string_params_or_400(schema=Schema())
            return response(data=params)

        return app

    def test_query_string_parameter_validation(self):
        resp = self.app.test_client().get(path="/stuffs?foo=one")
        expected = messages.query_string_validation_failed._asdict()
        expected["errors"] = {"foo": "Not a valid integer."}
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, expected)

        resp = self.app.test_client().get(path="/stuffs?bar=true")
        expected = messages.query_string_validation_failed._asdict()
        expected["errors"] = {"foo": "Missing data for required field."}

        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, expected)

        resp = self.app.test_client().get(path="/stuffs?foo=1&unexpected=true")
        expected = messages.query_string_validation_failed._asdict()
        expected["errors"] = {"unexpected": "Unknown field."}

        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json, expected)

        resp = self.app.test_client().get(path="/stuffs?foo=1&bar=true&baz=1,2,3")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"foo": 1, "bar": True, "baz": [1, 2, 3]})



"""
    Request Utilities
    ~~~~~~~~~~~~~~~~

    Utilities for request handlers.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import collections
import copy
from typing import Any, Dict, Iterator, List, Optional, Type, Union, overload

import marshmallow
from marshmallow import Schema
from flask import Response
from flask import jsonify
from flask import request
from werkzeug.datastructures import Headers
from werkzeug.exceptions import BadRequest as WerkzeugBadRequest

from flask_rebar import compat
from flask_rebar import errors
from flask_rebar import messages
from flask_rebar.utils.defaults import USE_DEFAULT
from flask_rebar.utils.marshmallow_objects_helpers import get_marshmallow_objects_schema


class HeadersProxy(collections.abc.Mapping):
    """
    Marshmallow expects objects being deserialized to be instances of `Mapping`.

    This wraps werkzeug's `EnvironHeaders` to ensure that they're an instance of `Mapping`.

    :param werkzeug.datastructures.EnvironHeaders headers:
    """

    __slots__ = ("headers",)

    def __init__(self, headers: Headers) -> None:
        self.headers = headers

    def __len__(self) -> int:
        return len(self.headers)

    def __iter__(self) -> Iterator[str]:
        # EnvironHeaders.__iter__ yields tuples of (key, value).
        # We want to mimic a dict and yield keys.
        return iter(self.headers.keys())

    def __contains__(self, item: Any) -> bool:
        return item in self.headers

    def __getitem__(self, key: Any) -> str:
        return self.headers[key]


def response(
    data: Optional[Any],
    status_code: int = 200,
    headers: Optional[Headers] = None,
    mimetype: Optional[str] = None,
) -> Response:
    """
    Constructs a flask.jsonify response.

    :param dict data: The JSON body of the response
    :param int status_code: HTTP status code to use in the response
    :param dict headers: Additional headers to attach to the response
    :param str mimetype: Default Content-Type response header
    :rtype: flask.Response
    """
    resp = jsonify(data) if data is not None else Response()

    resp.status_code = status_code

    if mimetype:
        if headers is not None:
            headers.update({"Content-Type": mimetype})
        else:
            headers = Headers({"Content-Type": mimetype})

    if headers is not None:
        response_headers = dict(resp.headers)
        response_headers.update(headers)
        resp.headers = Headers(response_headers)

    return resp


def marshal(data: Any, schema: Schema) -> Dict[str, Any]:
    """
    Dumps an object with the given marshmallow.Schema.

    :raises: marshmallow.ValidationError if the given data fails validation
      of the schema.
    """
    schema = normalize_schema(schema)

    return compat.dump(schema=schema, data=data)


@overload
def normalize_schema(schema: None) -> None:
    ...


@overload
def normalize_schema(schema: Type[USE_DEFAULT]) -> Type[USE_DEFAULT]:
    ...


@overload
def normalize_schema(schema: Union[Schema, Type[Schema]]) -> Schema:
    ...


def normalize_schema(
    schema: Any,
) -> Union[Schema, Type[Schema], Type[USE_DEFAULT], None]:
    """
    This allows for either an instance of a marshmallow.Schema or the class
    itself to be passed to functions.
    For Marshmallow-objects support, if a Model class is passed, return its __schema__

    Possible types:
    - schema instance -> return itself
    - schema class -> return instance of schema class
    - marshmallow-objects Model class -> return schema class (is this right?)
    - marshmallow-objects Model instance -> return schema class (is this right?)
    - None -> return None
    - USE_DEFAULT -> return USE_DEFAULT
    """
    if schema not in (None, USE_DEFAULT) and not isinstance(schema, marshmallow.Schema):
        # See if we were handed a marshmallow_objects Model class or instance:
        mo_schema = get_marshmallow_objects_schema(schema)
        if mo_schema:
            model = schema
            schema = mo_schema
            # If __swagger_title__ is defined on the Model, propagate that down:
            if hasattr(model, "__swagger_title__"):
                schema.__swagger_title__ = model.__swagger_title__
        else:
            # assume we were passed a Schema class (not an instance)
            schema = schema()
    return schema


def raise_400_for_marshmallow_errors(
    errs: Dict[str, Any], msg: Union[str, messages.ErrorMessage]
) -> errors.BadRequest:
    """
    Throws a 400 error properly formatted from the given marshmallow errors.

    :param dict: Error dictionary as returned by marshmallow
    :param Union[str,messages.ErrorMessage] msg: The overall message to use in the response.
    :raises: errors.BadRequest
    """
    if not errs:
        return errors.BadRequest(msg=msg)

    copied = copy.deepcopy(errs)

    _format_marshmallow_errors_for_response_in_place(copied)

    additional_data = {"errors": copied}

    return errors.BadRequest(msg=msg, additional_data=additional_data)


def get_json_body_params_or_400(schema: Schema) -> Dict[str, Any]:
    """
    Retrieves the JSON body of a request, validating/loading the payload
    with a given marshmallow.Schema.

    :param schema:
    :rtype: dict
    """
    body = _get_json_body_or_400()

    return _get_data_or_400(
        schema=schema, data=body, message=messages.body_validation_failed
    )


def get_query_string_params_or_400(schema: Schema) -> Dict[str, Any]:
    """
    Retrieves the query string of a request, validating/loading the parameters
    with a given marshmallow.Schema.

    :param schema:
    :rtype: dict
    """
    # Use the request.args MultiDict in case a validator wants to
    # do something with several of the same query param (e.g. ?foo=1&foo=2), in
    # which case it will need the getlist method
    query_multidict = request.args.copy()

    return _get_data_or_400(
        schema=schema,
        data=query_multidict,
        message=messages.query_string_validation_failed,
    )


def get_header_params_or_400(schema: Schema) -> Dict[str, Any]:
    schema = compat.exclude_unknown_fields(schema)
    return _get_data_or_400(
        schema=schema,
        data=HeadersProxy(request.headers),
        message=messages.header_validation_failed,
    )


def _get_data_or_400(
    schema: Schema, data: Any, message: messages.ErrorMessage
) -> Dict[str, Any]:
    schema = normalize_schema(schema)
    try:
        return compat.load(schema=schema, data=data)
    except marshmallow.ValidationError as e:
        raise raise_400_for_marshmallow_errors(errs=e.messages_dict, msg=message)


def _get_json_body_or_400() -> Union[List[Any], Dict[str, Any]]:
    """
    Retrieves the JSON payload of the current request, throwing a 400 error
    if the request doesn't include a valid JSON payload.
    """
    if "application/json" not in request.headers.get("content-type", ""):
        raise errors.BadRequest(messages.unsupported_content_type)

    if (not request.data) or (len(request.data) == 0):
        raise errors.BadRequest(messages.empty_json_body)

    try:
        body = request.get_json()
    except WerkzeugBadRequest:
        raise errors.BadRequest(messages.invalid_json)

    if not isinstance(body, list) and not isinstance(body, dict):
        # request.get_json_from_resp() treats strings as valid JSON, which is technically
        # true... but they're not valid objects. So let's throw an error on
        # primitive types.
        raise errors.BadRequest(messages.invalid_json)

    return body


def _format_marshmallow_errors_for_response_in_place(errs: Dict[str, Any]) -> None:
    """
    Reformats an error dictionary returned by marshmallow to an error
    dictionary we can send in a response.

    This transformation happens in place, so make sure to pass in a copy
    of the errors...
    """
    # These are errors on the entire schema, not a specific field
    # Let's rename these too something slightly less cryptic
    if "_schema" in errs:
        errs["_general"] = errs.pop("_schema")

    for field, value in errs.items():
        # In most cases we'll only have a single error for a field,
        # but marshmallow gives us a list regardless.
        # Let's try to reduce the complexity of the error response and convert
        # these lists to a single string.
        if isinstance(value, list) and len(value) == 1:
            errs[field] = value[0]
        elif isinstance(value, dict):
            # Recurse! Down the rabbit hole...
            _format_marshmallow_errors_for_response_in_place(value)



import enum

from flask_rebar.swagger_generation import swagger_words as sw
from werkzeug.routing import BaseConverter
from werkzeug.routing import ValidationError


class TodoType(str, enum.Enum):
    user = "user"
    group = "group"


class TodoTypeConverter(BaseConverter):
    def to_python(self, value):
        try:
            return TodoType(value)
        except ValueError:
            raise ValidationError()

    def to_url(self, obj):
        return obj.value

    @staticmethod
    def to_swagger():
        return {
            sw.type_: sw.string,
            sw.enum: [t.value for t in TodoType],
        }



"""
    Rebar Extension
    ~~~~~~~~~~~~~~~

    The main entry point for Flask-Rebar, including a Flask extension
    and a registry for request handlers.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import annotations
import sys
from collections import defaultdict
from collections import namedtuple
from copy import copy
from functools import wraps
from flask import current_app, g, jsonify, request, Response
from flask.app import Flask
from marshmallow import Schema
from typing import (
    overload,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException
from werkzeug.routing import RequestRedirect
from werkzeug.utils import find_modules
from werkzeug.utils import import_string

from flask_rebar import messages
from flask_rebar import errors
from flask_rebar.authenticators import Authenticator
from flask_rebar.errors import HttpJsonError
from flask_rebar.messages import ErrorMessage
from flask_rebar.utils.defaults import USE_DEFAULT
from flask_rebar.utils.request_utils import marshal
from flask_rebar.utils.request_utils import response
from flask_rebar.utils.request_utils import get_header_params_or_400
from flask_rebar.utils.request_utils import get_json_body_params_or_400
from flask_rebar.utils.request_utils import get_query_string_params_or_400
from flask_rebar.utils.request_utils import normalize_schema
from flask_rebar.utils.deprecation import deprecated, deprecated_parameters
from flask_rebar.swagger_generation.swagger_generator_base import SwaggerGenerator
from flask_rebar.swagger_generation.swagger_generator_v2 import SwaggerV2Generator
from flask_rebar.swagger_ui import create_swagger_ui_blueprint


MOVED_PERMANENTLY_ERROR = RequestRedirect
PERMANENT_REDIRECT_ERROR = RequestRedirect

# for type hinting decorators
P = ParamSpec("P")
T = TypeVar("T")

JsonType = Union[Dict[str, Any], List[Any], str, int, float, bool, None]
Data = Union[bytes, JsonType]


def _convert_authenticator_to_authenticators(
    authenticator: Optional[Union[Authenticator, Type[USE_DEFAULT]]]
) -> List[Union[Authenticator, Type[USE_DEFAULT]]]:
    if authenticator is None:
        return []
    elif isinstance(authenticator, Authenticator) or authenticator is USE_DEFAULT:
        return [authenticator]
    else:
        raise ValueError(
            "authenticator must be an instance of Authenticator, USE_DEFAULT, or None."
        )


def _unpack_view_func_return_value(
    rv: Union[
        Tuple[Data, int, Dict[str, str]],
        Tuple[Data, int],
        Tuple[Data, Dict[str, str]],
        Data,
    ],
) -> Tuple[
    Union[
        Tuple[Data, int, Dict[str, str]],
        Tuple[Data, int],
        Tuple[Data, Dict[str, str]],
        Data,
    ],
    int,
    Any,
]:
    """
    Normalize a return value from a view function into a tuple of (body, status, headers).

    This imitates Flask's own `Flask.make_response` method.

    :param rv: (body, status, headers), (body, status), (body, headers), or body
    :return: (body, status, headers)
    :rtype: tuple
    """
    headers: Any = {}
    data, status = rv, 200

    if isinstance(rv, tuple):
        if len(rv) == 3:
            data, status, headers = rv
        elif len(rv) == 2:
            if isinstance(rv[1], (Headers, dict, tuple, list)):
                data, headers = rv
            else:
                data, status = rv
        else:
            raise TypeError(
                "The view function did not return a valid response tuple."
                " The tuple must have the form (body, status, headers),"
                " (body, status), or (body, headers)."
            )

    return data, int(status), headers


def _wrap_handler(
    f: Callable[P, T],
    authenticators: Optional[List[Authenticator]] = None,
    query_string_schema: Optional[Schema] = None,
    request_body_schema: Optional[Schema] = None,
    headers_schema: Optional[Schema] = None,
    response_body_schema: Optional[Dict[int, Schema]] = None,
    mimetype: Optional[str] = None,
) -> Callable[P, Union[T, Response]]:
    """
    Wraps a handler function before registering it with a Flask application.

    :param f:
    :returns: a new, wrapped handler function
    """
    # authenticators can be a single Authenticator, a list of Authenticators, or None.
    if isinstance(authenticators, Authenticator):
        authenticators = [authenticators]

    @wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> Union[T, Response]:
        if authenticators:
            first_error = None
            for authenticator in authenticators:
                try:
                    authenticator.authenticate()
                    break  # Short-circuit on first successful authentication
                except (errors.Unauthorized, errors.Forbidden) as e:
                    first_error = first_error or e
            else:
                raise first_error or errors.Unauthorized

        if query_string_schema:
            g.validated_args = get_query_string_params_or_400(
                schema=query_string_schema
            )

        if request_body_schema:
            g.validated_body = get_json_body_params_or_400(schema=request_body_schema)

        if headers_schema:
            g.validated_headers = get_header_params_or_400(schema=headers_schema)

        rv: Any = f(*args, **kwargs)

        if not response_body_schema:
            return rv

        if isinstance(rv, current_app.response_class):
            schema = response_body_schema[rv.status_code]
            # The schema may be set to None to bypass marshaling (e.g. for 204 responses).
            if schema is None:
                return rv
            # Otherwise, ensure the response body conforms to the promised schema.
            schema.loads(rv.data)  # May raise ValidationError.
            return rv

        data, status_code, headers = _unpack_view_func_return_value(rv)
        schema = response_body_schema[status_code]  # May raise KeyError.

        # The schema may be set to None to bypass marshaling (e.g. for 204 responses).
        if schema is None:
            return response(
                data=data, status_code=status_code, headers=headers, mimetype=mimetype
            )

        marshaled = marshal(data=data, schema=schema)
        return response(
            data=marshaled, status_code=status_code, headers=headers, mimetype=mimetype
        )

    return wrapped


def get_validated_body() -> Dict[str, Any]:
    """
    Retrieve the result of validating/transforming an incoming request body with
    the `request_body_schema` a handler was registered with.

    :rtype: dict
    """
    return g.validated_body


def get_validated_args() -> Dict[str, Any]:
    """
    Retrieve the result of validating/transforming an incoming request's query
    string with the `query_string_schema` a handler was registered with.

    :rtype: dict
    """
    return g.validated_args


def get_validated_headers() -> Dict[str, str]:
    """
    Retrieve the result of validating/transforming an incoming request's headers
    with the `headers_schema` a handler was registered with.

    :rtype: dict
    """
    return g.validated_headers


@overload
def normalize_prefix(prefix: str) -> str:
    ...


@overload
def normalize_prefix(prefix: None) -> None:
    ...


def normalize_prefix(prefix: Optional[str]) -> Optional[str]:
    """
    Removes slashes from a URL path prefix.

    :param str prefix:
    :rtype: str
    """
    if prefix and prefix.startswith("/"):
        prefix = prefix[1:]
    if prefix and prefix.endswith("/"):
        prefix = prefix[:-1]

    return prefix


def prefix_url(prefix: str, url: str) -> str:
    """
    Returns a new URL with the prefix prepended to the provided URL.

    :param str prefix:
    :param str url:
    :rtype: str
    """
    prefix = normalize_prefix(prefix)
    url = url[1:] if url.startswith("/") else url
    return f"/{prefix}/{url}"


# Metadata about a declared handler function. This can be used to both
# declare the flask routing and to autogenerate swagger.
class PathDefinition(
    namedtuple(
        "_PathDefinition",
        [
            "func",
            "path",
            "method",
            "endpoint",
            "response_body_schema",
            "query_string_schema",
            "request_body_schema",
            "headers_schema",
            "authenticators",
            "tags",
            "mimetype",
            "hidden",
            "summary",
        ],
    )
):
    __slots__ = ()

    @deprecated_parameters(
        authenticator=(
            "authenticators",
            "3.0",
            _convert_authenticator_to_authenticators,
        )
    )
    def __new__(cls, *args: Any, **kwargs: Any) -> "PathDefinition":
        return super().__new__(cls, *args, **kwargs)

    @property
    @deprecated("authenticator", "3.0")
    def authenticator(self) -> Optional[Authenticator]:
        return self.authenticators[0] if self.authenticators else None


class HandlerRegistry:
    """
    Registry for request handlers.

    This should typically be instantiated via a :class:`Rebar` instance::

        rebar = Rebar()
        registry = rebar.create_handler_registry()

    Although it can be instantiated independently and added to the registry::

        rebar = Rebar()
        registry = HandlerRegistry()
        rebar.add_handler_registry(registry)

    :param str prefix:
        URL prefix for all handlers registered with this registry instance.
    :param Union(flask_rebar.authenticators.Authenticator, List(flask_rebar.authenticators.Authenticator), None)
        default_authenticators: List of Authenticators to use for all handlers as a default.
    :param marshmallow.Schema default_headers_schema:
        Schema to validate the headers on all requests as a default.
    :param str default_mimetype:
        Default response content-type if no content and not otherwise specified by the handler.
    :param flask_rebar.swagger_generation.swagger_generator.SwaggerGenerator swagger_generator:
        Object to generate a Swagger specification from this registry. This will be
        the Swagger generator that is used in the endpoints swagger and swagger UI
        that are added to the API.
        If left as None, a `SwaggerV2Generator` instance will be used.
    :param str spec_path:
        The Swagger specification as a JSON document will be hosted at this URL.
        If set as None, no swagger specification will be hosted.
    :param str spec_ui_path:
        The HTML Swagger UI will be hosted at this URL.
        If set as None, no Swagger UI will be hosted.
    :param list|str handlers:
         packages to search for modules where handlers are added for this registry.
    """

    @deprecated_parameters(
        default_authenticator=(
            "default_authenticators",
            "3.0",
            _convert_authenticator_to_authenticators,
        ),
        swagger_path="spec_path",  # we didn't specify an EOL for these two, maybe they just live forever (fine imho)
        swagger_ui_path="spec_ui_path",
    )
    def __init__(
        self,
        prefix: Optional[str] = None,
        default_authenticators: Union[Authenticator, List[Authenticator], None] = None,
        default_headers_schema: Optional[Schema] = None,
        default_mimetype: Optional[str] = None,
        swagger_generator: Optional[SwaggerGenerator] = None,
        spec_path: str = "/swagger",
        spec_ui_path: str = "/swagger/ui",
        handlers: Optional[Union[List[str], str]] = None,
    ) -> None:
        # default_authenticators can be a single Authenticator, a list of Authenticators, or None.
        if isinstance(default_authenticators, Authenticator):
            default_authenticators = [default_authenticators]
        elif default_authenticators is None:
            default_authenticators = []

        self.prefix = normalize_prefix(prefix)
        self._paths: Dict[str, Dict[str, PathDefinition]] = defaultdict(dict)
        self.default_authenticators = default_authenticators
        self.default_headers_schema = default_headers_schema
        self.default_mimetype = default_mimetype
        self.swagger_generator = swagger_generator or SwaggerV2Generator()
        self.spec_path = spec_path
        self.spec_ui_path = spec_ui_path
        if handlers is None:
            self.handlers: List[str] = []
        else:
            self.handlers = handlers if isinstance(handlers, list) else [handlers]

    @property
    @deprecated("default_authenticators", "3.0")
    def default_authenticator(self) -> Optional[Authenticator]:
        return self.default_authenticators[0] if self.default_authenticators else None

    def set_default_authenticator(self, authenticator: Authenticator) -> None:
        """
        Sets a handler authenticator to be used by default.

        :param Union(None, flask_rebar.authenticators.Authenticator) authenticator:
        """
        self.default_authenticators = (
            [authenticator] if authenticator is not None else []
        )

    def set_default_authenticators(self, authenticators: List[Authenticator]) -> None:
        """
        Sets the handler authenticators to be used by default.

        :param Union(List(flask_rebar.authenticators.Authenticator)) authenticators:
        """
        self.default_authenticators = authenticators or []

    def set_default_headers_schema(self, headers_schema: Schema) -> None:
        """
        Sets the schema to be used by default to validate incoming headers.

        :param marshmallow.Schema headers_schema:
        """
        self.default_headers_schema = normalize_schema(headers_schema)

    def clone(self) -> HandlerRegistry:
        """
        Returns a new, shallow-copied instance of :class:`HandlerRegistry`.

        :rtype: HandlerRegistry
        """
        return copy(self)

    def _prefixed(self, path: str) -> str:
        if self.prefix:
            return prefix_url(prefix=self.prefix, url=path)
        else:
            return path

    def _prefixed_spec_path(self) -> str:
        return self._prefixed(self.spec_path)

    def _prefixed_spec_ui_path(self) -> str:
        return self._prefixed(self.spec_ui_path)

    @property
    def paths(self) -> Dict[str, Dict[str, PathDefinition]]:
        # We duplicate the paths so we can modify the path definitions right before
        # they are accessed.
        paths: Dict[str, Dict[str, PathDefinition]] = defaultdict(dict)

        for path, methods in self._paths.items():
            for method, definition_ in methods.items():
                path = definition_.path

                if self.prefix:
                    path = prefix_url(prefix=self.prefix, url=path)

                paths[path][method] = PathDefinition(
                    func=definition_.func,
                    path=path,
                    method=definition_.method,
                    endpoint=definition_.endpoint,
                    response_body_schema=definition_.response_body_schema,
                    query_string_schema=definition_.query_string_schema,
                    request_body_schema=definition_.request_body_schema,
                    headers_schema=definition_.headers_schema,
                    authenticators=definition_.authenticators,
                    tags=definition_.tags,
                    mimetype=definition_.mimetype,
                    hidden=definition_.hidden,
                    summary=definition_.summary,
                )

        return paths

    @deprecated_parameters(
        authenticator=(
            "authenticators",
            "3.0",
            _convert_authenticator_to_authenticators,
        )
    )
    def add_handler(
        self,
        func: Callable,
        rule: str,
        method: str = "GET",
        endpoint: Optional[str] = None,
        response_body_schema: Optional[Dict[int, Schema]] = None,
        query_string_schema: Optional[Schema] = None,
        request_body_schema: Optional[Schema] = None,
        headers_schema: Union[Type[USE_DEFAULT], Schema] = USE_DEFAULT,
        authenticators: Union[
            Type[USE_DEFAULT], List[Authenticator], Authenticator
        ] = USE_DEFAULT,
        tags: Optional[Sequence[str]] = None,
        mimetype: Union[Type[USE_DEFAULT], str] = USE_DEFAULT,
        hidden: bool = False,
        summary: Optional[str] = None,
    ) -> None:
        """
        Registers a function as a request handler.

        :param func:
            The Flask "view_func"
        :param str rule:
            The Flask "rule"
        :param str method:
            The HTTP method this handler accepts
        :param str endpoint:
        :param dict[int, marshmallow.Schema] response_body_schema:
            Dictionary mapping response codes to schemas to use to marshal
            the response. For now this assumes everything is JSON.
        :param marshmallow.Schema query_string_schema:
            Schema to use to deserialize query string arguments.
        :param marshmallow.Schema request_body_schema:
            Schema to use to deserialize the request body. For now this
            assumes everything is JSON.
        :param Type[USE_DEFAULT]|None|marshmallow.Schema headers_schema:
            Schema to use to grab and validate headers.
        :param Type[USE_DEFAULT]|None|List(Authenticator)|Authenticator authenticators:
            A list of authenticator objects to authenticate incoming requests.
            If left as USE_DEFAULT, the Rebar's default will be used.
            Set to None to make this an unauthenticated handler.
        :param Sequence[str] tags:
            Arbitrary strings to tag the handler with. These will translate to Swagger operation tags.
        :param Type[USE_DEFAULT]|None|str mimetype:
            Content-Type header to add to the response schema
        :param bool hidden:
            if hidden, documentation is not created for this request handler by default
        :param str summary:
        """
        # Fix #115: if we were passed bare classes we'll go ahead and instantiate
        headers_schema = normalize_schema(headers_schema)
        request_body_schema = normalize_schema(request_body_schema)
        query_string_schema = normalize_schema(query_string_schema)
        if response_body_schema:
            # Ensure we wrap in appropriate default (200) dict if we were passed a single Schema or class:
            if not isinstance(response_body_schema, Mapping):
                response_body_schema = {200: response_body_schema}
            # use normalize_schema to convert any class reference(s) to instantiated schema(s):
            response_body_schema = {
                code: normalize_schema(schema)
                for (code, schema) in response_body_schema.items()
            }

        # authenticators can be a list of Authenticators, a single Authenticator, USE_DEFAULT, or None
        authenticators_list: Sequence[Union[Type[USE_DEFAULT], Authenticator]] = []
        if isinstance(authenticators, list):
            authenticators_list = authenticators
        elif isinstance(authenticators, Authenticator) or authenticators is USE_DEFAULT:
            authenticators_list = [authenticators]
        elif authenticators is None:
            authenticators_list = []

        self._paths[rule][method] = PathDefinition(
            func=func,
            path=rule,
            method=method,
            endpoint=endpoint,
            response_body_schema=response_body_schema,
            query_string_schema=query_string_schema,
            request_body_schema=request_body_schema,
            headers_schema=headers_schema,
            authenticators=authenticators_list,
            tags=tags,
            mimetype=mimetype,
            hidden=hidden,
            summary=summary,
        )

    @deprecated_parameters(
        authenticator=(
            "authenticators",
            "3.0",
            _convert_authenticator_to_authenticators,
        )
    )
    def handles(
        self,
        rule: str,
        method: str = "GET",
        endpoint: Optional[str] = None,
        response_body_schema: Optional[Dict[int, Schema]] = None,
        query_string_schema: Optional[Schema] = None,
        request_body_schema: Optional[Schema] = None,
        headers_schema: Union[Type[USE_DEFAULT], Schema] = USE_DEFAULT,
        authenticators: Union[
            Type[USE_DEFAULT], List[Authenticator], Authenticator
        ] = USE_DEFAULT,
        tags: Optional[Sequence[str]] = None,
        mimetype: Union[Type[USE_DEFAULT], str] = USE_DEFAULT,
        hidden: bool = False,
        summary: Optional[str] = None,
    ) -> Callable:
        """
        Same arguments as :meth:`HandlerRegistry.add_handler`, except this can
        be used as a decorator.
        """

        def wrapper(f: Callable[P, T]) -> Callable[P, T]:
            self.add_handler(
                func=f,
                rule=rule,
                method=method,
                endpoint=endpoint,
                response_body_schema=response_body_schema,
                query_string_schema=query_string_schema,
                request_body_schema=request_body_schema,
                headers_schema=headers_schema,
                authenticators=authenticators,
                tags=tags,
                mimetype=mimetype,
                hidden=hidden,
                summary=summary,
            )
            return f

        return wrapper

    def register(self, app: Flask) -> None:
        self._register_routes(app=app)
        self._register_swagger(app=app)
        self._register_swagger_ui(app=app)

    def _register_routes(self, app: Flask) -> None:
        for handler in self.handlers:
            for handler_mod in find_modules(import_path=handler, recursive=True):
                import_string(handler_mod)

        for path, methods in self.paths.items():
            for method, definition_ in methods.items():
                if definition_.endpoint:
                    endpoint = definition_.endpoint
                else:
                    endpoint = definition_.func.__name__

                if self.prefix:
                    endpoint = ".".join((self.prefix, endpoint))

                authenticators = []
                for authenticator in definition_.authenticators:
                    if authenticator is USE_DEFAULT:
                        authenticators.extend(self.default_authenticators)
                    else:
                        authenticators.append(authenticator)

                app.add_url_rule(
                    rule=definition_.path,
                    view_func=_wrap_handler(
                        f=definition_.func,
                        authenticators=authenticators,
                        query_string_schema=definition_.query_string_schema,
                        request_body_schema=definition_.request_body_schema,
                        headers_schema=(
                            self.default_headers_schema
                            if definition_.headers_schema is USE_DEFAULT
                            else definition_.headers_schema
                        ),
                        response_body_schema=definition_.response_body_schema,
                        mimetype=(
                            self.default_mimetype
                            if definition_.mimetype is USE_DEFAULT
                            else definition_.mimetype
                        ),
                    ),
                    methods=[definition_.method],
                    endpoint=endpoint,
                )

    def _register_swagger(self, app: Flask) -> None:
        swagger_endpoint = "get_swagger"

        if self.prefix:
            swagger_endpoint = ".".join((self.prefix, swagger_endpoint))

        if self.spec_path:

            @app.route(
                self._prefixed_spec_path(), methods=["GET"], endpoint=swagger_endpoint
            )
            def get_swagger() -> Response:
                swagger = self.swagger_generator.generate_swagger(
                    registry=self, host=request.host_url.rstrip("/")
                )
                return response(data=swagger)

    def _register_swagger_ui(self, app: Flask) -> None:
        blueprint_name = "swagger_ui"

        if self.prefix:
            blueprint_name = self.prefix.replace(".", "_") + blueprint_name

        if self.spec_ui_path:
            blueprint = create_swagger_ui_blueprint(
                name=blueprint_name,
                ui_url=self._prefixed_spec_ui_path(),
                swagger_url=self._prefixed_spec_path(),
            )
            app.register_blueprint(
                blueprint=blueprint, url_prefix=self._prefixed_spec_ui_path()
            )


class Rebar:
    """
    The main entry point for the Flask-Rebar extension.

    This registers handler registries with the Flask application and initializes
    all the Flask-Rebar goodies.

    Example usage::

        app = Flask(__name__)
        rebar = Rebar()
        registry = rebar.create_handler_registry()

        @registry.handles()
        def handler():
            ...

        rebar.init_app(app)

    """

    def __init__(self) -> None:
        self.handler_registries: Set[HandlerRegistry] = set()
        self.paths: Dict[str, Dict[str, PathDefinition]] = defaultdict(dict)
        self.uncaught_exception_handlers: List[Callable] = []
        # If a developer doesn't wish to advertise that they are using rebar this can be used to control
        # the name of the attribute in error responses, or set to None to suppress inclusion of error codes entirely
        self.error_code_attr = "rebar_error_code"
        self.validate_on_dump = False

    @deprecated_parameters(
        default_authenticator=(
            "default_authenticators",
            "3.0",
            _convert_authenticator_to_authenticators,
        )
    )
    def create_handler_registry(
        self,
        prefix: Optional[str] = None,
        default_authenticators: Union[List[Authenticator], Authenticator, None] = None,
        default_headers_schema: Optional[Schema] = None,
        default_mimetype: Optional[str] = None,
        swagger_generator: Optional[SwaggerGenerator] = None,
        spec_path: str = "/swagger",
        swagger_ui_path: str = "/swagger/ui",
        handlers: Optional[List[str]] = None,
    ) -> HandlerRegistry:
        """
        Create a new handler registry and add to this extension's set of
        registered registries.

        When calling :meth:`Rebar.init_app`, all registries created via this method
        will be registered with the Flask application.

        Parameters are the same for the :class:`HandlerRegistry` constructor.

        :param str prefix:
            URL prefix for all handlers registered with this registry instance.
        :param Union(List(Authenticator), Authenticator, None) default_authenticators:
            List of Authenticators to use for all handlers as a default.
        :param marshmallow.Schema default_headers_schema:
            Schema to validate the headers on all requests as a default.
        :param str default_mimetype:
            Default response content-type if no content and not otherwise specified by the handler.
        :param flask_rebar.swagger_generation.swagger_generator.SwaggerGeneratorI swagger_generator:
            Object to generate a Swagger specification from this registry. This will be
            the Swagger generator that is used in the endpoints swagger and swagger UI
            that are added to the API.
            If left as None, a `SwaggerV2Generator` instance will be used.
        :param str spec_path:
            The OpenAPI specification as a JSON document will be hosted at this URL.
            If set as None, no swagger specification will be hosted.
        :param str swagger_ui_path:
            The HTML Swagger UI will be hosted at this URL.
            If set as None, no Swagger UI will be hosted.
        :param list handlers: directories where handlers should be imported from.
        :rtype: HandlerRegistry
        """
        registry = HandlerRegistry(
            prefix=prefix,
            default_authenticators=default_authenticators,
            default_headers_schema=default_headers_schema,
            default_mimetype=default_mimetype,
            swagger_generator=swagger_generator,
            spec_path=spec_path,
            spec_ui_path=swagger_ui_path,
            handlers=handlers,
        )
        self.add_handler_registry(registry=registry)
        return registry

    def add_handler_registry(self, registry: HandlerRegistry) -> None:
        """
        Register a handler registry with the extension.

        There is no need to call this if a handler registry was created
        via :meth:`Rebar.create_handler_registry`.

        :param HandlerRegistry registry:
        """
        self.handler_registries.add(registry)

    @property
    def validated_body(self) -> Dict[str, Any]:
        """
        Proxy to the result of validating/transforming an incoming request body with
        the `request_body_schema` a handler was registered with.

        :rtype: dict
        """
        return get_validated_body()

    @property
    def validated_args(self) -> Dict[str, Any]:
        """
        Proxy to the result of validating/transforming an incoming request's query
        string with the `query_string_schema` a handler was registered with.

        :rtype: dict
        """
        return get_validated_args()

    @property
    def validated_headers(self) -> Dict[str, str]:
        """
        Proxy to the result of validating/transforming an incoming request's headers
        with the `headers_schema` a handler was registered with.

        :rtype: dict
        """
        return get_validated_headers()

    def add_uncaught_exception_handler(self, func: Callable) -> None:
        """
        Add a function that will be called for uncaught exceptions, i.e. exceptions
        that will result in a 500 error.

        This function should accept the exception instance as a single positional argument.

        All handlers will be called in the order they are added.

        :param Callable func:
        """
        self.uncaught_exception_handlers.append(func)

    def init_app(self, app: Flask) -> None:
        """
        Register all the handler registries with a Flask application.

        :param flask.Flask app:
        """
        self._init_error_handling(app=app)

        for registry in self.handler_registries:
            registry.register(app=app)

        app.extensions["rebar"] = {
            "instance": self,
            "handler_registries": self.handler_registries,
        }

    def _init_error_handling(self, app: Flask) -> None:
        @app.errorhandler(errors.HttpJsonError)
        def handle_http_error(error: HttpJsonError) -> Response:
            return self._create_json_error_response(
                message=error.error_message,
                http_status_code=error.http_status_code,
                additional_data=error.additional_data,
            )

        @app.errorhandler(400)
        @app.errorhandler(404)
        @app.errorhandler(405)
        def handle_werkzeug_http_error(error: HTTPException) -> Response:
            return self._create_json_error_response(
                message=error.description, http_status_code=error.code
            )

        @app.errorhandler(MOVED_PERMANENTLY_ERROR)
        @app.errorhandler(PERMANENT_REDIRECT_ERROR)
        def handle_request_redirect_error(error: RequestRedirect) -> Response:
            return self._create_json_error_response(
                message=error.name,
                http_status_code=error.code,
                additional_data={"new_url": error.new_url},
                headers={"Location": error.new_url},
            )

        def run_unhandled_exception_handlers(exception: BaseException) -> None:
            exc_info = sys.exc_info()
            current_app.log_exception(exc_info=exc_info)

            for func in self.uncaught_exception_handlers:
                func(exception)

        @app.errorhandler(Exception)
        def handle_generic_error(error: Exception) -> Response:
            run_unhandled_exception_handlers(error)

            if current_app.debug:
                raise error
            else:
                return self._create_json_error_response(
                    message=messages.internal_server_error, http_status_code=500
                )

        @app.teardown_request
        def teardown(exception: Optional[BaseException]) -> None:
            if isinstance(exception, SystemExit):
                try:
                    run_unhandled_exception_handlers(exception)
                except Exception:
                    # make sure the exception handlers dont prevent teardown
                    pass

    def _create_json_error_response(
        self,
        message: Optional[Union[ErrorMessage, str]],
        http_status_code: Optional[int],
        additional_data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        """
        Compiles a response object for an error.

        :param Union[messages.ErrorMessage,str] message:
        :param int http_status_code:
          An optional, application-specific error code to add to the response.
        :param dict additional_data:
          Additional JSON data to attach to the response.
        :param dict headers:
          Additional headers to attach to the response.
        :rtype: flask.Response
        """
        if isinstance(message, messages.ErrorMessage):
            message_text = message.message
            rebar_error_code = message.rebar_error_code
        else:
            message_text = message
            rebar_error_code = None
        body = {"message": message_text}
        if additional_data:
            body.update(additional_data)
        if rebar_error_code and self.error_code_attr:
            body[self.error_code_attr] = rebar_error_code
        resp = jsonify(body)
        if headers:
            for key, value in headers.items():
                resp.headers[key] = value
        if http_status_code is not None:
            resp.status_code = http_status_code
        return resp



"""
    General-Purpose Utilities
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Utilities that are not specific to request-handling (you'll find those in request_utils.py).

    :copyright: Copyright 2019 Autodesk, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import annotations
import functools
import sys
import warnings
from typing import Any, Callable, Dict, NamedTuple, Optional, Tuple, TypeVar, Union

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

from werkzeug.local import LocalProxy as module_property  # noqa


# ref http://jtushman.github.io/blog/2014/05/02/module-properties/ for background on
# use of werkzeug LocalProxy to simulate "module properties"
# end result: singleton config can be accessed by, e.g.,
#    from flask_rebar.utils.deprecation import config as deprecation_config
#    deprecation_config.warning_type = YourFavoriteWarning


T = TypeVar("T")
P = ParamSpec("P")


class DeprecationConfig:
    """
    Singleton class to allow one-time set of deprecation config controls
    """

    __instance = None

    @staticmethod
    def getInstance() -> DeprecationConfig:
        """Static access method."""
        if DeprecationConfig.__instance is None:
            return DeprecationConfig()
        return DeprecationConfig.__instance

    def __init__(self) -> None:
        """Virtually private constructor."""
        if DeprecationConfig.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            DeprecationConfig.__instance = self
            self.warning_type = FutureWarning


@module_property
def config() -> DeprecationConfig:
    return DeprecationConfig.getInstance()


def deprecated(
    new_func: Optional[Union[str, Tuple[str, str]]] = None,
    eol_version: Optional[str] = None,
) -> Callable:
    """
    :param Union[str, (str, str)] new_func: Name (or name and end-of-life version) of replacement
    :param str eol_version: Version in which this function may no longer work
    :return:
    Raise a deprecation warning for decorated function.
    Tuple is supported for new_func just in case somebody infers it as an option based on the way we
    deprecate params..
    If tuple form is used for new_func AND eol_version is provided, eol_version will trump whatever is
    found in the tuple; caveat emptor
    """

    def decorator(f: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            new, eol, _ = _validated_deprecation_spec(new_func)
            eol = eol_version or eol
            _deprecation_warning(f.__name__, new, eol, stacklevel=3)
            return f(*args, **kwargs)

        return wrapper

    return decorator


def deprecated_parameters(**aliases: Any) -> Callable:
    """
    Adapted from https://stackoverflow.com/a/49802489/977046
    :param aliases: Keyword args in the form {old_param_name = Union[new_param_name, (new_param_name, eol_version),
                    (new_param_name, eol_version, coerce_func)]}
                    where eol_version is the version in which the alias may case to be recognized and coerce_func is a
                    function used to coerce old values to new values.
    :return: function decorator that will apply aliases to param names and raise DeprecationWarning
    """

    def decorator(f: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            new_kwargs = _remap_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **new_kwargs)

        return wrapper

    return decorator


class DeprecationSpec(NamedTuple):
    new_name: Optional[str]
    eol_version: Optional[str]
    coerce_func: Optional[Callable]


def _validated_deprecation_spec(
    spec: Optional[Union[str, Tuple[Any, ...]]]
) -> DeprecationSpec:
    """
    :param Union[new_name, (new_name, eol_version), (new_name, eol_version, coerce_func)] spec:
        new name and/or expected end-of-life version
    :return: (str new_name, str eol_version, func coerce_func),
        normalized to tuple and sanitized to deal with malformed inputs
    Parse a deprecation spec (string or tuple) to a standardized namedtuple form.
    If spec is provided as a bare value (presumably string), we'll treat as new name with no end-of-life version
    If spec is provided (likely on accident) as a 1-element tuple, we'll treat same as a bare value
    If spec is provided as a tuple with more than 3 elements, we'll simply ignore the extraneous
    """
    new_name = None
    eol_version = None
    coerce_func = None
    if type(spec) is tuple:
        if len(spec) > 0:
            new_name = str(spec[0]) if spec[0] else None
        if len(spec) > 1:
            eol_version = str(spec[1]) if spec[1] else None
        if len(spec) > 2:
            coerce_func = spec[2]
    elif spec:
        new_name = str(spec)
    validated = DeprecationSpec(new_name, eol_version, coerce_func)
    return validated


def _remap_kwargs(
    func_name: str, kwargs: Dict[str, Any], aliases: Dict[str, str]
) -> Dict[str, Any]:
    """
    Adapted from https://stackoverflow.com/a/49802489/977046
    """
    remapped_args = dict(kwargs)
    for alias, new_spec in aliases.items():
        if alias in remapped_args:
            new, eol_version, coerce_func = _validated_deprecation_spec(new_spec)
            if new in remapped_args:
                raise TypeError(f"{func_name} received both {alias} and {new}")
            else:
                _deprecation_warning(alias, new, eol_version, stacklevel=4)
                if new:
                    value = remapped_args.pop(alias)
                    if coerce_func is not None:
                        value = coerce_func(value)
                    remapped_args[new] = value
    return remapped_args


def _deprecation_warning(
    old_name: str,
    new_name: Optional[str],
    eol_version: Optional[str],
    stacklevel: int = 1,
) -> None:
    eol_clause = f" and may be removed in version {eol_version}" if eol_version else ""
    replacement_clause = f"; use {new_name}" if new_name else ""
    msg = f"{old_name} is deprecated{eol_clause}{replacement_clause}"
    warnings.warn(message=msg, category=config.warning_type, stacklevel=stacklevel)  # type: ignore



import os
import subprocess
import json
import contextlib


this_directory = os.path.dirname(os.path.realpath(__file__))
todo_app_filepath = os.path.join(this_directory, "todo.py")
todo_output_filepath = os.path.join(this_directory, "todo_output.md")


@contextlib.contextmanager
def app():
    print("Starting app...")

    app_process = None
    try:
        app_process = subprocess.Popen(
            args=["python", todo_app_filepath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Wait for the app to startup by polling the service
        up = False
        while not up:
            try:
                subprocess.check_output(
                    "curl -s http://127.0.0.1:5000/swagger", shell=True
                )
                up = True
            except subprocess.SubprocessError:
                pass

        yield

    finally:
        if app_process:
            app_process.terminate()


def main():
    output = []

    with app():
        output.extend(
            [
                "# cURL and examples/todo.py",
                "Here's a snippet of playing with the application inside todo.py.",
                "",
            ]
        )

        for title, commands in [
            ("Swagger for free!", ["curl -s -XGET http://127.0.0.1:5000/swagger"]),
            (
                "Request validation!",
                [
                    'curl -s -XPATCH http://127.0.0.1:5000/todos/1 -H "X-MyApp-Key: my-api-key" -H "Content-Type: application/json" -d \'{"complete": "wrong type, for demonstration of validation"}\''
                ],
            ),
            (
                "Authentication!",
                [
                    "curl -s -XGET http://127.0.0.1:5000/todos",
                    'curl -s -XGET http://127.0.0.1:5000/todos -H "X-MyApp-Key: my-api-key"',
                ],
            ),
            (
                "CRUD!",
                [
                    'curl -s -XPOST http://127.0.0.1:5000/todos -H "X-MyApp-Key: my-api-key" -H "Content-Type: application/json" -d \'{"complete": false, "description": "Find product market fit"}\'',
                    'curl -s -XPATCH http://127.0.0.1:5000/todos/1 -H "X-MyApp-Key: my-api-key" -H "Content-Type: application/json" -d \'{"complete": true}\'',
                    'curl -s -XGET http://127.0.0.1:5000/todos -H "X-MyApp-Key: my-api-key"',
                ],
            ),
        ]:
            output.extend([title, "```"])

            for command in commands:
                print(command)

                result = subprocess.check_output(command, shell=True)

                output.extend(
                    ["$ " + command, json.dumps(json.loads(result), indent=2)]
                )

            output.extend(["```", ""])

    print(f"Writing output to {todo_output_filepath}")

    with open(todo_output_filepath, "w") as f:
        f.write("\n".join(output))

    print("Done!")


if __name__ == "__main__":
    main()



# Rebar relies heavily on Marshmallow.
# These schemas will be used to validate incoming data, marshal outgoing
# data, and to automatically generate a Swagger specification.

from flask_rebar.validation import RequestSchema, ResponseSchema
from flask_rebar.swagger_generation.marshmallow_to_swagger import EnumField
from marshmallow import fields, pre_dump, pre_load

from .converters import TodoType


class CreateTodoSchema(RequestSchema):
    complete = fields.Boolean(required=True)
    description = fields.String(required=True)
    type = EnumField(TodoType, load_default=TodoType.user)


class UpdateTodoSchema(CreateTodoSchema):
    # This schema provides an example of one way to re-use another schema while making some fields optional
    # a "partial" schema in Marshmallow parlance:
    def __init__(self, **kwargs):
        super_kwargs = dict(kwargs)
        partial_arg = super_kwargs.pop("partial", True)
        # Note: if you only want to mark some fields as partial, pass partial= a collection of field names, e.g.,:
        # partial_arg = super_kwargs.pop('partial', ('description', ))
        super().__init__(partial=partial_arg, **super_kwargs)


class GetTodoListSchema(RequestSchema):
    complete = fields.Boolean()


class TodoSchema(ResponseSchema):
    id = fields.Integer(required=True)
    complete = fields.Boolean(required=True)
    description = fields.String(required=True)
    type = EnumField(TodoType, required=True)


class TodoResourceSchema(ResponseSchema):
    data = fields.Nested(TodoSchema)

    @pre_dump
    @pre_load
    def envelope_in_data(self, data, **kwargs):
        if type(data) is not dict or "data" not in data.keys():
            return {"data": data}
        else:
            return data


class TodoListSchema(ResponseSchema):
    data = fields.Nested(TodoSchema, many=True)

    @pre_dump
    @pre_load
    def envelope_in_data(self, data, **kwargs):
        if type(data) is not dict or "data" not in data.keys():
            return {"data": data}
        else:
            return data
